from copy import copy, deepcopy

import pytest
from lxml import etree

from translate.misc.multistring import multistring
from translate.storage import aresource
from translate.storage.base import TranslationStore

from . import test_monolingual


class TestAndroidResourceUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = aresource.AndroidResourceUnit

    def __check_escape(self, string, xml, target_language=None) -> None:
        """Helper that checks that a string is output with the right escape."""
        unit = self.UnitClass("teststring")

        if target_language is not None:
            store = TranslationStore()
            store.settargetlanguage(target_language)
            unit._store = store

        unit.target = string

        print("unit.target:", repr(unit.target))
        print("xml:", repr(xml))

        assert str(unit) == xml

    def __check_parse(self, string, xml) -> None:
        """Helper that checks that a string is parsed correctly."""
        parser = etree.XMLParser(strip_cdata=False)

        translatable = 'translatable="false"' not in xml
        et = etree.fromstring(xml, parser)
        unit = self.UnitClass.createfromxmlElement(et)

        print("unit.target:", repr(unit.target))
        print("string:", string)
        print("translatable:", repr(unit.istranslatable()))

        assert unit.target == string
        assert unit.istranslatable() == translatable

    ############################ Check string escape ##########################

    def test_escape_message_with_newline(self) -> None:
        string = "message\nwith newline"
        xml = '<string name="teststring">message\\nwith newline</string>\n'
        self.__check_escape(string, xml)

    def test_escape_quotes_with_newline(self) -> None:
        string = "'message'\nwith newline"
        xml = "<string name=\"teststring\">\\'message\\'\\nwith newline</string>\n"
        self.__check_escape(string, xml)

    def test_escape_message_with_newline_in_xml(self) -> None:
        string = "message\n\nwith newline in xml\n"
        xml = '<string name="teststring">message\\n\\nwith newline in xml\\n</string>\n'
        self.__check_escape(string, xml)

    def test_escape_twitter(self) -> None:
        string = "@twitterescape"
        xml = '<string name="teststring">\\@twitterescape</string>\n'
        self.__check_escape(string, xml)

    def test_escape_quote(self) -> None:
        string = "quote 'escape'"
        xml = "<string name=\"teststring\">quote \\'escape\\'</string>\n"
        self.__check_escape(string, xml)

    def test_escape_question(self) -> None:
        string = "?question?"
        xml = '<string name="teststring">\\?question?</string>\n'
        self.__check_escape(string, xml)

    def test_escape_double_space(self) -> None:
        string = "double  space"
        xml = '<string name="teststring">"double  space"</string>\n'
        self.__check_escape(string, xml)

    def test_escape_leading_space(self) -> None:
        string = " leading space"
        xml = '<string name="teststring">" leading space"</string>\n'
        self.__check_escape(string, xml)

    def test_escape_tailing_space(self) -> None:
        string = "tailing space "
        xml = '<string name="teststring">"tailing space "</string>\n'
        self.__check_escape(string, xml)

    def test_escape_xml_entities(self) -> None:
        string = ">xml&entities"
        xml = '<string name="teststring">&gt;xml&amp;entities</string>\n'
        self.__check_escape(string, xml)

    def test_escape_html_code(self) -> None:
        string = "some <b>html code</b> here"
        xml = '<string name="teststring">some <b>html code</b> here</string>\n'
        self.__check_escape(string, xml)

    def test_escape_html_code_quote(self) -> None:
        string = "some <b>html code</b> 'here'"
        xml = "<string name=\"teststring\">some <b>html code</b> \\'here\\'</string>\n"
        self.__check_escape(string, xml)

    def test_escape_html_code_quote_newline(self) -> None:
        string = "some \n<b>html code</b> 'here'"
        xml = (
            "<string name=\"teststring\">some \\n<b>html code</b> \\'here\\'</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_arrows(self) -> None:
        string = "<<< arrow"
        xml = '<string name="teststring">&lt;&lt;&lt; arrow</string>\n'
        self.__check_escape(string, xml)

    def test_escape_link(self) -> None:
        string = '<a href="http://example.net">link</a>'
        xml = (
            '<string name="teststring">\n'
            '  <a href="http://example.net">link</a>\n'
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_link_and_text(self) -> None:
        string = '<a href="http://example.net">link</a> and text'
        xml = (
            '<string name="teststring"><a href="http://example.net">link'
            "</a> and text</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_blank_string(self) -> None:
        string = ""
        xml = '<string name="teststring"></string>\n'
        self.__check_escape(string, xml)

    def test_plural_escape_message_with_newline(self) -> None:
        mString = multistring(
            ["one message\nwith newline", "other message\nwith newline"]
        )
        xml = (
            '<plurals name="teststring">\n'
            '    <item quantity="one">one message\\nwith newline</item>'
            '<item quantity="other">other message\\nwith newline</item>'
            "</plurals>\n"
        )
        self.__check_escape(mString, xml, "en")

    def test_plural_invalid_lang(self) -> None:
        mString = multistring(["one message", "other message"])
        xml = (
            '<plurals name="teststring">\n'
            '    <item quantity="one">one message</item>'
            '<item quantity="other">other message</item>'
            "</plurals>\n"
        )
        self.__check_escape(mString, xml, "invalid")

    def test_escape_html_quote(self) -> None:
        string = "start 'here' <b>html code 'to escape'</b> also 'here'"
        xml = (
            "<string name=\"teststring\">start \\'here\\' <b>html code \\'to escape\\'</b> also \\'here\\'"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_leading_space(self) -> None:
        string = " <b>html code 'to escape'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"> <b>html code \\'to escape\\'</b> some \\'here\\'"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_trailing_space(self) -> None:
        string = "<b>html code 'to escape'</b> some 'here' "
        xml = (
            "<string name=\"teststring\"><b>html code \\'to escape\\'</b> some \\'here\\' "
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_with_ampersand(self) -> None:
        string = "<b>html code 'to escape'</b> some 'here' with &amp; char"
        xml = (
            "<string name=\"teststring\"><b>html code \\'to escape\\'</b> some \\'here\\' with &amp; char"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_double_space(self) -> None:
        string = "<b>html code 'to  escape'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"><b>html code \\'to \\u0020escape\\'</b> some \\'here\\'"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_deep_double_space(self) -> None:
        string = "<b>html code 'to  <i>escape</i>'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"><b>html code \\'to \\u0020<i>escape</i>\\'</b> some \\'here\\'"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_complex_xml(self) -> None:
        string = '<g:test xmlns:g="ttt" g:somevalue="aaaa  &quot;  aaa">value</g:test> &amp; outer &gt; <br/>text'
        xml = (
            '<string name="teststring">'
            '<g:test xmlns:g="ttt" g:somevalue="aaaa  &quot;  aaa">value</g:test> &amp; outer &gt; <br/>text'
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_quoted_newlines(self) -> None:
        self.__check_escape(
            "\n\nstring with newlines",
            r"""<string name="teststring">\n\nstring with newlines</string>
""",
        )

    ############################ Check string parse ###########################

    def test_parse_message_with_newline(self) -> None:
        string = "message\nwith newline"
        xml = '<string name="teststring">message\\nwith newline</string>\n'
        self.__check_parse(string, xml)

    def test_parse_message_with_newline_in_xml(self) -> None:
        string = "message \nwith\n newline\n in xml"
        xml = (
            '<string name="teststring">message\n\\nwith\\n\nnewline\\n\nin xml'
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_twitter(self) -> None:
        string = "@twitterescape"
        xml = '<string name="teststring">\\@twitterescape</string>\n'
        self.__check_parse(string, xml)

    def test_parse_question(self) -> None:
        string = "?"
        xml = '<string name="question">\\?</string>'
        self.__check_parse(string, xml)

    def test_parse_quote(self) -> None:
        string = "quote 'escape'"
        xml = "<string name=\"teststring\">quote \\'escape\\'</string>\n"
        self.__check_parse(string, xml)

    def test_parse_double_space(self) -> None:
        string = "double  space"
        xml = '<string name="teststring">"double  space"</string>\n'
        self.__check_parse(string, xml)

    def test_parse_leading_space(self) -> None:
        string = " leading space"
        xml = '<string name="teststring">" leading space"</string>\n'
        self.__check_parse(string, xml)

    def test_parse_quoted_newlines(self) -> None:
        self.__check_parse(
            "\n\n\n\nstring with newlines",
            r"""<string name="teststring">"
\n
\nstring with newlines"</string>
""",
        )

    def test_parse_xml_entities(self) -> None:
        string = ">xml&entities"
        xml = '<string name="teststring">&gt;xml&amp;entities</string>\n'
        self.__check_parse(string, xml)

    def test_parse_html_code(self) -> None:
        string = "some <b>html code</b> here"
        xml = '<string name="teststring">some <b>html code</b> here</string>\n'
        self.__check_parse(string, xml)

    def test_parse_escaped_html_code_flag(self) -> None:
        xml = '<string name="teststring">some &lt;b&gt;html code&lt;/b&gt; here</string>\n'
        parser = etree.XMLParser(strip_cdata=False)
        unit = self.UnitClass.createfromxmlElement(etree.fromstring(xml, parser))

        assert unit.target == "some <b>html code</b> here"
        assert unit.target_markup_mode == unit.TARGET_MARKUP_ESCAPED

    def test_parse_escaped_angle_text_flag(self) -> None:
        xml = '<string name="teststring">&lt;not set&gt;</string>\n'
        parser = etree.XMLParser(strip_cdata=False)
        unit = self.UnitClass.createfromxmlElement(etree.fromstring(xml, parser))

        assert unit.target == "<not set>"
        assert unit.target_markup_mode == unit.TARGET_MARKUP_PLAIN

    def test_parse_nested_html_code_flag(self) -> None:
        xml = '<string name="teststring">some <b>html code</b> here</string>\n'
        parser = etree.XMLParser(strip_cdata=False)
        unit = self.UnitClass.createfromxmlElement(etree.fromstring(xml, parser))

        assert unit.target == "some <b>html code</b> here"
        assert unit.target_markup_mode == unit.TARGET_MARKUP_XML

    def test_parse_cdata_flag(self) -> None:
        xml = (
            '<string name="teststring"><![CDATA[<a href="%1$s">Terms</a>]]></string>\n'
        )
        parser = etree.XMLParser(strip_cdata=False)
        unit = self.UnitClass.createfromxmlElement(etree.fromstring(xml, parser))

        assert unit.target == '<a href="%1$s">Terms</a>'
        assert unit.target_markup_mode == unit.TARGET_MARKUP_CDATA

    def test_parse_arrows(self) -> None:
        string = "<<< arrow"
        xml = '<string name="teststring">&lt;&lt;&lt; arrow</string>\n'
        self.__check_parse(string, xml)

    def test_parse_link(self) -> None:
        string = '<a href="http://example.net">link</a>'
        xml = (
            '<string name="teststring"><a href="http://example.net">link</a></string>\n'
        )
        self.__check_parse(string, xml)

    def test_parse_link_and_text(self) -> None:
        string = '<a href="http://example.net">link</a> and text'
        xml = (
            '<string name="teststring"><a href="http://example.net">link'
            "</a> and text</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_blank_string(self) -> None:
        string = ""
        xml = '<string name="teststring"></string>\n'
        self.__check_parse(string, xml)

    def test_parse_trailing_space(self) -> None:
        string = "test"
        xml = '<string name="teststring">test </string>\n'
        self.__check_parse(string, xml)

    def test_parse_trailing_spaces(self) -> None:
        string = "test"
        xml = '<string name="teststring">test    </string>\n'
        self.__check_parse(string, xml)

    def test_parse_trailing_escaped_newline(self) -> None:
        string = "test\n"
        xml = '<string name="teststring">test\\n</string>\n'
        self.__check_parse(string, xml)

    def test_parse_leading_escaped_newline(self) -> None:
        string = "\ntest"
        xml = '<string name="teststring">\\ntest</string>\n'
        self.__check_parse(string, xml)

    def test_parse_leading_spaces(self) -> None:
        string = "test"
        xml = '<string name="teststring">    test</string>\n'
        self.__check_parse(string, xml)

    def test_parse_trailing_newline(self) -> None:
        string = "test"
        xml = '<string name="teststring">test\n</string>\n'
        self.__check_parse(string, xml)

    def test_parse_many_quotes(self) -> None:
        string = "test"
        xml = '<string name="teststring">""""""""""test"""""""</string>\n'
        self.__check_parse(string, xml)

    def test_parse_blank_string_again(self) -> None:
        string = ""
        xml = '<string name="teststring"/>\n'
        self.__check_parse(string, xml)

    def test_parse_double_quotes_string(self) -> None:
        """Check that double quotes got removed."""
        string = "double quoted text"
        xml = '<string name="teststring">"double quoted text"</string>\n'
        self.__check_parse(string, xml)

    def test_parse_newline_in_string(self) -> None:
        """
        Check that newline is read as space.

        At least it seems to be what Android does.
        """
        string = "newline\nin string"
        xml = '<string name="teststring">newline\\nin string</string>\n'
        self.__check_parse(string, xml)

    def test_parse_not_translatable_string(self) -> None:
        string = "string"
        xml = '<string name="teststring" translatable="false">string</string>\n'
        self.__check_parse(string, xml)

    def test_plural_parse_message_with_newline(self) -> None:
        mString = multistring(
            ["one message\nwith newline", "other message\nwith newline"]
        )
        xml = (
            '<plurals name="teststring">\n'
            '    <item quantity="one">one message\\nwith newline</item>\n'
            '    <item quantity="other">other message\\nwith newline</item>\n'
            "</plurals>\n"
        )
        self.__check_parse(mString, xml)

    def test_plural_parse_message_with_comments(self) -> None:
        mString = multistring(["one message", "other message"])
        xml = (
            '<plurals name="teststring">\n'
            "    <!-- comment of one string -->\n"
            '    <item quantity="one">one message</item>\n'
            "    <!-- comment of other string -->\n"
            '    <item quantity="other">other message</item>\n'
            "</plurals>\n"
        )
        self.__check_parse(mString, xml)

    def test_parse_html_quote(self) -> None:
        string = "start 'here' <b>html code 'to escape'</b> also 'here'"
        xml = (
            "<string name=\"teststring\">start \\'here\\' <b>html code \\'to escape\\'</b> also \\'here\\'"
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_leading_space(self) -> None:
        string = "<b>html code 'to escape'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"> <b>html code \\'to escape\\'</b> some \\'here\\'"
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_leading_space_quoted(self) -> None:
        string = " <b>html code 'to escape'</b> some 'here'"
        xml = (
            '<string name="teststring">" "<b>"html code \'to escape\'"</b>" some \'here\'"'
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_trailing_space(self) -> None:
        string = "<b>html code 'to escape'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"><b>html code \\'to escape\\'</b> some \\'here\\' "
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_trailing_space_quoted(self) -> None:
        string = "<b>html code 'to escape'</b> some 'here' "
        xml = (
            '<string name="teststring"><b>"html code \'to escape\'"</b>" some \'here\' "'
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_with_ampersand(self) -> None:
        string = "<b>html code 'to escape'</b> some 'here' with &amp; char"
        xml = (
            "<string name=\"teststring\"><b>html code \\'to escape\\'</b> some \\'here\\' with &amp; char"
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_double_space_quoted(self) -> None:
        string = "<b>html code 'to  escape'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"><b>html code 'to \\u0020escape'</b> some 'here'"
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_deep_double_space_quoted(self) -> None:
        string = "<b>html code 'to  <i>  escape</i>'</b> some 'here'"
        xml = (
            '<string name="teststring"><b>"html code \'to  "<i>"  escape"</i>\\\'</b> some \\\'here\\\''
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_complex_xml(self) -> None:
        string = '<g:test xmlns:g="ttt" g:somevalue="aaaa  &quot;  aaa">value</g:test> outer &amp; text'
        xml = (
            '<string name="teststring">'
            '<g:test xmlns:g="ttt" g:somevalue="aaaa  &quot;  aaa">value</g:test> outer &amp; text'
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_unicode(self) -> None:
        self.__check_parse("test", r'<string name="test">\utest</string>')
        self.__check_parse("\u0230", r'<string name="test">\u0230</string>')

    def test_single_unescaped(self) -> None:
        string = "a b c d"
        xml = '<string name="teststring">a\nb\tc d</string>'
        self.__check_parse(string, xml)

    def test_single_escaped_alone(self) -> None:
        string = "a\nb\tc d"
        xml = '<string name="teststring">a"\n"b"\t"c" "d</string>'
        self.__check_parse(string, xml)

    def test_single_escaped_full(self) -> None:
        string = "a\nb\tc d"
        xml = '<string name="teststring">"a\nb\tc d"</string>'
        self.__check_parse(string, xml)

    def test_parse_quoted_escaped_percent(self) -> None:
        self.__check_parse(
            """You are about to remove "%s" from your library""",
            r"""<string name="remove_manga">You are about to remove \"%s\" from your library</string>""",
        )

    def test_parse_quoted_percent(self) -> None:
        self.__check_parse(
            """You are about to remove %s from your library""",
            """<string name="remove_manga">You are about to remove "%s" from your library</string>""",
        )

    def test_parse_quoted_quote(self) -> None:
        self.__check_parse(
            """You are about to remove "%s" from your library""",
            r"""<string name="remove_manga">You are about to "remove \"%s\" from" your library</string>""",
        )

    def test_parse_unparied_quote(self) -> None:
        self.__check_parse(
            "You can view it here soon: %s",
            """<string name="cmd_camera_response_success">"You can view it here soon: %s</string>""",
        )
        self.__check_parse(
            "You can view it here soon: %s",
            """<string name="cmd_camera_response_success">You can view it here soon: %s"</string>""",
        )

    def test_parse_escape_slash(self) -> None:
        self.__check_parse(
            """Přechod na novější verzi databáze účtu \\<g id="account">%s</g>\\""",
            r"""<string name="upgrade_database_format">Přechod na novější verzi databáze účtu \\<g id="account">%s</g>\\</string>""",
        )

    def test_parse_escape_ignored(self) -> None:
        self.__check_parse(
            """Přechod na novější verzi databáze účtu \\<g id="account">%s</g>""",
            r"""<string name="upgrade_database_format">Přechod na novější verzi databáze účtu \\<g id="account">%s</g>\x</string>""",
        )

    def test_parse_escaped_quote_end(self) -> None:
        self.__check_parse(
            'No file %1$s found in archive "%2$s"',
            r"""<string name="restore_backup_file_not_found">No file %1$s found in archive \"%2$s\"</string>""",
        )

    def test_parse_quoted_xml(self) -> None:
        self.__check_parse(
            """Your phone needs to use Bluetooth to securely collect and share IDs with other phones that are nearby.

<g example="Corona-Warn">%1$s</g> can notify you if you were exposed to someone who reported to be diagnosed positive.

The date, duration, and signal strength associated with an exposure will be shared with the app.""",
            """<string name="exposure_confirm_start_summary">"Your phone needs to use Bluetooth to securely collect and share IDs with other phones that are nearby.

<g example="Corona-Warn">%1$s</g> can notify you if you were exposed to someone who reported to be diagnosed positive.

The date, duration, and signal strength associated with an exposure will be shared with the app."</string>""",
        )


class TestAndroidResourceFile(test_monolingual.TestMonolingualStore):
    StoreClass = aresource.AndroidResourceFile

    def test_targetlanguage_default_handlings(self) -> None:
        store = self.StoreClass()

        # Initial value is None
        assert store.gettargetlanguage() is None

        # sourcelanguage shouldn't change the targetlanguage
        store.setsourcelanguage("en")
        assert store.gettargetlanguage() is None

        # targetlanguage setter works correctly
        store.settargetlanguage("de")
        assert store.gettargetlanguage() == "de"

        # explicit targetlanguage wins over filename
        store.filename = "dommy/values-it/res.xml"
        assert store.gettargetlanguage() == "de"

    def test_targetlanguage_auto_detection_filename(self) -> None:
        store = self.StoreClass()

        # Check language auto_detection
        store.filename = "project/values-it/res.xml"
        assert store.gettargetlanguage() == "it"

    def test_targetlanguage_auto_detection_filename_default_language(self) -> None:
        store = self.StoreClass()

        store.setsourcelanguage("en")

        # Check language auto_detection
        store.filename = "project/values/res.xml"
        assert store.gettargetlanguage() == "en"

    def test_targetlanguage_auto_detection_invalid_filename(self) -> None:
        store = self.StoreClass()

        store.setsourcelanguage("en")

        store.filename = "project/invalid_directory/res.xml"
        assert store.gettargetlanguage() is None

        store.filename = "invalid_directory"
        assert store.gettargetlanguage() is None

    def test_namespaces(self) -> None:
        content = """<resources xmlns:tools="http://schemas.android.com/tools">
            <string name="string1" tools:ignore="PluralsCandidate">string1</string>
            <string name="string2">string2</string>
        </resources>"""
        store = self.StoreClass()
        store.parse(content)
        newstore = self.StoreClass()
        newstore.addunit(store.units[0], new=True)
        print(newstore)
        assert b'<resources xmlns:tools="http://schemas.android.com/tools">' in bytes(
            newstore
        )

    def test_serialize(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="test">Test</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content

    def test_add_formatting(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="test">Test</string>
</resources>"""
        other = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="other">Test</string>
    <string name="other2">Test</string>
</resources>"""
        expected = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="test">Test</string>
    <string name="other">Test</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)
        otherstore = self.StoreClass()
        otherstore.parse(other)
        store.addunit(otherstore.units[0], True)
        assert bytes(store) == expected

    def test_entity(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE resources [
<!ENTITY appName "Linphone">
]>
<resources>
    <string name="app_name">&appName;</string>
    <string name="app_core">&appName; Core</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)
        assert store.units[0].source == "&appName;"
        store.units[0].target = "&appName;"
        assert store.units[1].source == "&appName; Core"
        store.units[1].target = "&appName; Core"
        assert bytes(store) == content

    def test_invalid_entity(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE resources [
<!ENTITY appName "Linphone">
]>
<resources>
    <string name="app_name">&appName;</string>
    <string name="app_core">&appName; Core</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content.encode())
        store.units[0].target = "&appName;"
        store.units[1].target = "&otherName; Core"
        assert bytes(store).decode() == content.replace(
            "&appName; Core", "&amp;otherName; Core"
        )

    def test_quoted_entity_doctype_preserved(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE resources [
<!ENTITY quoted "a&quot;b&#10;c">
]>
<resources>
    <string name="quoted_value">&quoted;</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)
        store.units[0].target = "&quoted;"
        assert bytes(store) == content

    def test_indent(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE resources [
<!ENTITY url_privacy_policy "http://example.com/">
]>
<resources>
    <string name="privacy_policy"><u><a href="&url_privacy_policy;">Datenschutzerklärung</a></u></string>
</resources>
""".encode()
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content

    def test_edit_plural_markup(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den</item>
        <item quantity="few"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny</item>
        <item quantity="other"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "cs"
        store.parse(content)
        store.units[0].target = multistring(
            [
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu',
            ]
        )
        # Additional entry for many (decimal) is present
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den</item>
        <item quantity="few"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny</item>
        <item quantity="many"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu</item>
        <item quantity="other"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu</item>
    </plurals>
</resources>
"""
        )

    def test_parse_decimal_plurals(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den</item>
        <item quantity="few"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny</item>
        <item quantity="many"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dni</item>
        <item quantity="other"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu</item>
    </plurals>
</resources>
"""
        # Resulting parsed plurals are still the same
        store = self.StoreClass()
        store.targetlanguage = "cs"
        store.parse(content)
        assert store.units[0].target == multistring(
            [
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu',
            ]
        )

    def test_edit_to_plurals(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="teststring">Test %d string</string>
</resources>
"""
        # Resulting parsed plurals are still the same
        store = self.StoreClass()
        store.targetlanguage = "cs"
        store.parse(content.encode())
        assert store.units[0].target == "Test %d string"
        store.units[0].target = multistring(
            [
                "Test %d string",
                "Test %d strings",
                "Test %d strings",
            ]
        )
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Test %d string</item>
        <item quantity="few">Test %d strings</item>
        <item quantity="many">Test %d strings</item>
        <item quantity="other">Test %d strings</item>
    </plurals>
</resources>
"""
        )

    def entity_add(self, *, edit: bool) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE resources [
<!ENTITY appName "Zkouška">
]>
<resources>
    <string name="app_name">&appName;</string>
    <string name="app_core">&appName; Core</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content.encode())
        second = self.StoreClass()
        newunit = deepcopy(store.units[0])
        second.addunit(newunit, True)
        if edit:
            newunit.target = "&appName;"
        newunit = deepcopy(store.units[1])
        second.addunit(newunit, True)
        if edit:
            newunit.target = "&appName; Core"
        # The original store should be unchanged
        assert bytes(store).decode() == content
        # The new store should have same content
        assert bytes(second).decode() == content

    def test_entity_add_edit(self) -> None:
        self.entity_add(edit=True)

    def test_entity_add_noedit(self) -> None:
        self.entity_add(edit=False)

    def test_markup_remove(self) -> None:
        template = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="privacy_policy"><u>Datenschutzerklärung</u></string>
</resources>
"""
        content = template.encode("utf-8")
        newcontent = template.replace("<u>", "").replace("</u>", "").encode("utf-8")
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content
        store.units[0].target = "Datenschutzerklärung"
        assert bytes(store) == newcontent

    def test_markup_set(self) -> None:
        template = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <!--Multimedia tab-->
    <string name="id">Multimedia tab</string>
</resources>
"""
        content = template.encode("utf-8")
        newcontent = template.replace(">Multimedia tab<", ">Other <b>tab</b><").encode(
            "utf-8"
        )
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content
        store.units[0].target = "Other <b>tab</b>"
        assert bytes(store) == newcontent

    def test_markup_roundtrip_preserves_escaped_html(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other &lt;b&gt;tab&lt;/b&gt;</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        assert store.units[0].target == "Other <b>tab</b>"
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_ESCAPED

        store.units[0].target = "Next <b>tab</b>"
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Next &lt;b&gt;tab&lt;/b&gt;</string>
</resources>
"""
        )

    def test_namespaced_markup_roundtrip_preserves_escaped_xml(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">
    <string name="id">&lt;xliff:g&gt;%s&lt;/xliff:g&gt; den</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        assert store.units[0].target == "<xliff:g>%s</xliff:g> den"
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_ESCAPED

        store.units[0].target = store.units[0].target
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">
    <string name="id">&lt;xliff:g&gt;%s&lt;/xliff:g&gt; den</string>
</resources>
"""
        )

    def test_plain_angle_text_noop_edit_roundtrip(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">&lt;not set&gt;</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        assert store.units[0].target == "<not set>"
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_PLAIN

        store.units[0].target = store.units[0].target
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">&lt;not set&gt;</string>
</resources>
"""
        )

    def test_plain_angle_text_literal_entities_do_not_double_escape(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">&lt;not set&gt;</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_XML
        store.units[0].target = "&lt;not set&gt;"
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_PLAIN
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">&lt;not set&gt;</string>
</resources>
"""
        )

    def test_plain_angle_text_edit_stays_plain(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">&lt;not set&gt;</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        store.units[0].target = "<b>x</b>"

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_PLAIN
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">&lt;b&gt;x&lt;/b&gt;</string>
</resources>
"""
        )

    def test_target_markup_mode_setter_overrides_to_xml_markup(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other &lt;b&gt;tab&lt;/b&gt;</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_XML
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_XML
        store.units[0].target = "Next <b>tab</b>"

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_XML
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Next <b>tab</b></string>
</resources>
"""
        )

    def test_target_markup_mode_setter_overrides_to_plain(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other <b>tab</b></string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_PLAIN
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_PLAIN
        store.units[0].target = "Next <b>tab</b>"

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_PLAIN
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Next &lt;b&gt;tab&lt;/b&gt;</string>
</resources>
"""
        )

    def test_target_markup_deprecated_getter_alias(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other &lt;b&gt;tab&lt;/b&gt;</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        with pytest.deprecated_call(match="target_markup"):
            assert store.units[0].target_markup is True

        store = self.StoreClass()
        store.parse(
            b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other <b>tab</b></string>
</resources>
"""
        )

        with pytest.deprecated_call(match="target_markup"):
            assert store.units[0].target_markup is False

    def test_target_markup_deprecated_setter_alias_to_xml_markup(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other &lt;b&gt;tab&lt;/b&gt;</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        with pytest.deprecated_call(match="target_markup"):
            store.units[0].target_markup = False
        store.units[0].target = "Next <b>tab</b>"

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_XML
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Next <b>tab</b></string>
</resources>
"""
        )

    def test_target_markup_deprecated_setter_alias_to_escaped_markup(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other <b>tab</b></string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        with pytest.deprecated_call(match="target_markup"):
            store.units[0].target_markup = True
        store.units[0].target = "Next <b>tab</b>"

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_ESCAPED
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Next &lt;b&gt;tab&lt;/b&gt;</string>
</resources>
"""
        )

    def test_target_markup_mode_setter_overrides_to_escaped_markup(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other <b>tab</b></string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_ESCAPED
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_ESCAPED
        store.units[0].target = "Next <b>tab</b>"

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_ESCAPED
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Next &lt;b&gt;tab&lt;/b&gt;</string>
</resources>
"""
        )

    def test_cdata_roundtrip_preserves_existing_sections(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">By continuing <![CDATA[<a href="%1$s">Terms</a>]]> and <![CDATA[<a href="%2$s">Privacy</a>]]></string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_CDATA

        store.units[0].target = store.units[0].target
        assert bytes(store) == content

    def test_cdata_roundtrip_decodes_android_escapes(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id"><![CDATA[\\@twitterescape\\nwith newline]]></string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        assert store.units[0].target == "@twitterescape\nwith newline"
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_CDATA

        store.units[0].target = store.units[0].target
        assert bytes(store) == content

    def test_target_markup_mode_setter_overrides_to_cdata(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other tab</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_CDATA
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_CDATA
        store.units[0].target = 'Next <a href="%1$s">tab</a>'

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_CDATA
        assert store.units[0].target == 'Next <a href="%1$s">tab</a>'
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id"><![CDATA[Next <a href="%1$s">tab</a>]]></string>
</resources>
"""
        )

    def test_target_markup_mode_setter_to_cdata_applies_android_escaping(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other tab</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_CDATA
        store.units[0].target = "?attr/foo"

        assert store.units[0].target == "?attr/foo"
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_CDATA
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id"><![CDATA[\\?attr/foo]]></string>
</resources>
"""
        )

    def test_cdata_mode_setter_overrides_to_plain(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id"><![CDATA[<a href="%1$s">Terms</a>]]></string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_PLAIN
        store.units[0].target = '<a href="%1$s">Terms</a>'

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_PLAIN
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">&lt;a href=\\"%1$s\\"&gt;Terms&lt;/a&gt;</string>
</resources>
"""
        )

    def test_cdata_mode_setter_overrides_to_xml_markup(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id"><![CDATA[<a href="%1$s">Terms</a>]]></string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_XML
        store.units[0].target = '<a href="%1$s">Terms</a>'

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_XML
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id"><a href="%1$s">Terms</a></string>
</resources>
"""
        )

    def test_plural_markup_roundtrip_preserves_escaped_html(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Other &lt;b&gt;tab&lt;/b&gt;</item>
        <item quantity="other">Other &lt;b&gt;tabs&lt;/b&gt;</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "en"
        store.parse(content)

        assert store.units[0].target == multistring(
            ["Other <b>tab</b>", "Other <b>tabs</b>"]
        )
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_ESCAPED

        store.units[0].target = multistring(["Next <b>tab</b>", "Next <b>tabs</b>"])
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Next &lt;b&gt;tab&lt;/b&gt;</item>
        <item quantity="other">Next &lt;b&gt;tabs&lt;/b&gt;</item>
    </plurals>
</resources>
"""
        )

    def test_plural_target_markup_mode_setter_overrides_to_xml_markup(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Other &lt;b&gt;tab&lt;/b&gt;</item>
        <item quantity="other">Other &lt;b&gt;tabs&lt;/b&gt;</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "en"
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_XML
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_XML
        store.units[0].target = multistring(["Next <b>tab</b>", "Next <b>tabs</b>"])

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_XML
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Next <b>tab</b></item>
        <item quantity="other">Next <b>tabs</b></item>
    </plurals>
</resources>
"""
        )

    def test_plural_target_markup_mode_setter_overrides_to_escaped_markup(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Other <b>tab</b></item>
        <item quantity="other">Other <b>tabs</b></item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "en"
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_ESCAPED
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_ESCAPED
        store.units[0].target = multistring(["Next <b>tab</b>", "Next <b>tabs</b>"])

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_ESCAPED
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Next &lt;b&gt;tab&lt;/b&gt;</item>
        <item quantity="other">Next &lt;b&gt;tabs&lt;/b&gt;</item>
    </plurals>
</resources>
"""
        )

    def test_plural_markup_roundtrip_preserves_cdata(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><![CDATA[<a href="%1$s">Tab</a>]]></item>
        <item quantity="other"><![CDATA[<a href="%1$s">Tabs</a>]]></item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "en"
        store.parse(content)

        assert store.units[0].target == multistring(
            ['<a href="%1$s">Tab</a>', '<a href="%1$s">Tabs</a>']
        )
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_CDATA

        store.units[0].target = store.units[0].target
        assert bytes(store) == content

    def test_plural_target_markup_mode_setter_overrides_to_cdata(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Other tab</item>
        <item quantity="other">Other tabs</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "en"
        store.parse(content)

        store.units[0].target_markup_mode = store.units[0].TARGET_MARKUP_CDATA
        store.units[0].target = multistring(
            ['<a href="%1$s">Tab</a>', '<a href="%1$s">Tabs</a>']
        )

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_CDATA
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><![CDATA[<a href="%1$s">Tab</a>]]></item>
        <item quantity="other"><![CDATA[<a href="%1$s">Tabs</a>]]></item>
    </plurals>
</resources>
"""
        )

    def test_mixed_plural_cdata_noop_edit_keeps_target_values(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><![CDATA[<a href="%1$s">Tab</a>]]></item>
        <item quantity="other">Other tabs</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "en"
        store.parse(content)

        expected = multistring(['<a href="%1$s">Tab</a>', "Other tabs"])
        assert store.units[0].target == expected
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_CDATA

        store.units[0].target = store.units[0].target

        assert store.units[0].target == expected
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><![CDATA[<a href="%1$s">Tab</a>]]></item>
        <item quantity="other"><![CDATA[Other tabs]]></item>
    </plurals>
</resources>
"""
        )

    def test_plural_xml_markup_downgrades_to_plain(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Other <b>tab</b></item>
        <item quantity="other">Other <b>tabs</b></item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "en"
        store.parse(content)

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_XML

        store.units[0].target = multistring(["Next tab", "Next tabs"])

        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_PLAIN
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one">Next tab</item>
        <item quantity="other">Next tabs</item>
    </plurals>
</resources>
"""
        )

    def test_target_markup_mode_setter_rejects_invalid_value(self) -> None:
        store = self.StoreClass()
        store.parse(
            b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Value</string>
</resources>
"""
        )

        with pytest.raises(ValueError, match="Unsupported target markup mode"):
            store.units[0].target_markup_mode = "unsupported"

    def test_edit_plural_others(self) -> None:
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den</item>
        <item quantity="few"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny</item>
        <item quantity="many"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu</item>
        <item quantity="other"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "ru"
        store.parse(content)
        store.units[0].target = multistring(
            [
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu',
            ]
        )
        assert bytes(store) == content

    def test_markup_quotes_set(self) -> None:
        template = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Test</string>
</resources>
"""
        content = template.encode()
        newcontent = template.replace(
            ">Test<", ">Test <b>string</b> with \\u0020space<"
        )
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content
        store.units[0].target = "Test <b>string</b> with  space"
        assert bytes(store).decode() == newcontent

    def test_xliff_g(self) -> None:
        template = """<?xml version="1.0" encoding="utf-8"?>
<resources>
</resources>"""
        original = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Test: <xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">%s</xliff:g></string>
</resources>
"""
        expected = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other: <xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">%s</xliff:g></string>
</resources>
"""
        origstore = self.StoreClass()
        origstore.parse(original.encode())
        store = self.StoreClass()
        store.parse(template.encode())
        store.addunit(origstore.units[0], new=True)
        assert bytes(store).decode() == original
        store.units[
            0
        ].target = """Other: <xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">%s</xliff:g>"""
        assert bytes(store).decode() == expected

    def test_xliff_namespace(self) -> None:
        original = """<?xml version="1.0" encoding="utf-8"?>
<resources xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">
    <string name="id">Test</string>
</resources>
"""
        expected = """<?xml version="1.0" encoding="utf-8"?>
<resources xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">
    <string name="id">Test: <xliff:g>%s</xliff:g></string>
</resources>
"""
        store = self.StoreClass()
        store.parse(original.encode())
        assert bytes(store).decode() == original
        store.units[
            0
        ].target = """Test: <xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">%s</xliff:g>"""
        # The namespace should be flattened as it is defined in the toplevel element
        assert bytes(store).decode() == expected

    def test_edit_plural_zh_hk(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="one">%d visitor</item>
        <item quantity="other">%d visitors</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "zh-rHK"
        store.parse(content.encode())
        store.units[0].target = "%d 訪客"
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="other">%d 訪客</item>
    </plurals>
</resources>
"""
        )

    def test_edit_plural_b_zh_hk(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="one">%d visitor</item>
        <item quantity="other">%d visitors</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "b+zh+Hant+HK"
        store.parse(content.encode())
        store.units[0].target = "%d 訪客"
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="other">%d 訪客</item>
    </plurals>
</resources>
"""
        )

    def test_missing_plural(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="one">%d visitor</item>
        <item quantity="other">%d visitors</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "fr"
        store.parse(content.encode())
        assert store.units[0].target == multistring(["%d visitor", "", "%d visitors"])

    def test_partial_plural_roundtrip(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="one">%d visitor</item>
        <item quantity="other">%d visitors</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "fr"
        store.parse(content.encode())

        assert store.units[0].target == multistring(["%d visitor", "", "%d visitors"])

        store.units[0].target = store.units[0].target

        assert bytes(store).decode() == content

    def test_partial_plural_setter_omits_empty_translations(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="vms_num_visitors">%d visitors</string>
</resources>
"""
        store = self.StoreClass()
        store.targetlanguage = "fr"
        store.parse(content.encode())

        store.units[0].target = multistring(["%d visiteur", "", "%d visiteurs"])

        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="one">%d visiteur</item>
        <item quantity="other">%d visiteurs</item>
    </plurals>
</resources>
"""
        )

    def test_empty_missing_plural_tag(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="1" />
    <plurals name="2">
      <item quantity="other">%1$d tiles left in pool.</item>
    </plurals>
    <plurals name="3">
      <item quantity="other">%1$d tiles left in pool.</item>
      <item quantity="one">%1$d tile left in pool.</item>
    </plurals>
</resources>
"""
        store = self.StoreClass()
        store.parse(content.encode())
        assert store.units[0].target == multistring(["", ""])
        assert store.units[1].target == multistring(["", "%1$d tiles left in pool."])
        assert store.units[2].target == multistring(
            ["%1$d tile left in pool.", "%1$d tiles left in pool."]
        )

    def test_removeunit(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <!-- First -->
    <string name="test1">Test</string>
    <!-- Second -->
    <string name="test2">Test</string>
    <!-- Third -->
    <string name="test3">Test2</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content.encode())
        assert len(store.units) == 3
        assert bytes(store).decode() == content
        store.removeunit(store.units[1])
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <!-- First -->
    <string name="test1">Test</string>
    <!-- Third -->
    <string name="test3">Test2</string>
</resources>
"""
        )

    def test_cdata(self) -> None:
        body = r"""<Data>XXX<b>x</b><![CDATA[
    <html><head>
      <style type=\"text/css\">
        body { margin: 0px; }
      </style></head><body>
      <div style=\"background-color: transparent; color: %1$s; padding:
%2$dpx;\">
        <p>This app requires <strong>permission to access all
files</strong> on the storage.</p>
      </div>
    </body></html>
    ]]></Data>"""
        content = f"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="dialog_storage_permission_info" formatted="false">{body}</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content.encode())
        assert len(store.units) == 1
        assert store.units[0].target == body
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_XML
        assert bytes(store).decode() == content
        store.units[0].target = body
        assert bytes(store).decode() == content

    def test_cdata_text(self) -> None:
        body = """By continuing you accept our <![CDATA[<a href="%1$s">Terms of Use</a>]]> and <![CDATA[<a href="%2$s">Privacy Policy</a>]]>"""
        content = f"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="service_terms_agreement_notice">{body}</string>
</resources>
"""

        # Verify round trip
        store = self.StoreClass()
        store.parse(content.encode())
        assert len(store.units) == 1
        assert store.units[0].target == body
        assert store.units[0].target_markup_mode == store.units[0].TARGET_MARKUP_CDATA
        assert bytes(store).decode() == content
        store.units[0].target = body
        assert bytes(store).decode() == content

        # Verify copying unit to new store
        newstore = self.StoreClass()
        newstore.addunit(copy(store.units[0]), new=True)
        assert bytes(newstore).decode() == content
        assert (
            newstore.units[0].target_markup_mode
            == newstore.units[0].TARGET_MARKUP_CDATA
        )
        newstore.units[0].target = body
        assert bytes(newstore).decode() == content

        # Verify creating unit from scratch
        addstore = self.StoreClass()
        unit = addstore.UnitClass(body)
        unit.setid("service_terms_agreement_notice")
        unit.target = body
        addstore.addunit(unit, new=True)
        assert len(addstore.units) == 1
        assert addstore.units[0].target == body
        assert (
            addstore.units[0].target_markup_mode == addstore.units[0].TARGET_MARKUP_XML
        )
        assert bytes(addstore).decode() == content
        addstore.units[0].target = body
        assert bytes(addstore).decode() == content

    def test_prefix(self) -> None:
        body = "&lt; <b>body</b>"
        content = f"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="test">{body}</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content.encode())
        assert len(store.units) == 1
        assert store.units[0].target == body
        assert bytes(store).decode() == content

    def test_rtl(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="calendar_permission_required"><i>{app_name}</i> צרכה רשות לגשת ליומן שלך, על מנת ליצור תוכניות עבור עסקאות עתידיות חוזרות.</string>
</resources>
"""
        store = self.StoreClass()
        store.parse(content.encode())
        assert len(store.units) == 1
        assert (
            store.units[0].target
            == "<i>{app_name}</i> צרכה רשות לגשת ליומן שלך, על מנת ליצור תוכניות עבור עסקאות עתידיות חוזרות."
        )

    def test_tail(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="one"><b>One</b></string>in
    <string name="two"><b>Two</b></string>
</resources>"""
        store = self.StoreClass()
        store.parse(content.encode())
        assert len(store.units) == 2
        assert store.units[0].target == "<b>One</b>"
        assert store.units[1].target == "<b>Two</b>"

    def test_translatable_marking(self) -> None:
        data = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <!-- First -->
    <string name="test1" translateble="true">Test</string>
    <!-- Second -->
    <string name="test2">Test</string>
    <!-- Third -->
    <string name="test3" translatable="false">Test2</string>
</resources>"""
        store = self.StoreClass()
        store.parse(data.encode())
        store.units[1].marktranslatable(False)
        assert store.units[1].getid() == "test2"
        assert not store.units[1].istranslatable()
        store.units[1].marktranslatable(True)
        assert store.units[1].istranslatable()

        store.units[2].marktranslatable(True)
        assert store.units[2].getid() == "test3"
        assert store.units[2].istranslatable()

        store.units[0].marktranslatable(False)
        assert store.units[0].getid() == "test1"
        assert not store.units[0].istranslatable()

    def test_escaping(self) -> None:
        content = r"""<?xml version="1.0" encoding="utf-8" standalone="no"?>
<resources>
<string name="hello" weblate-flags="c-format, max-length:100">Hello, world!\n</string>
</resources>"""
        store = self.StoreClass()
        store.parse(content.encode())
        assert len(store.units) == 1
        assert store.units[0].target == "Hello, world!\n"


class TestMOKOResourceUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = aresource.MOKOResourceUnit

    def test_escape_quote(self) -> None:
        """Helper that checks that a string is output with the right escape."""
        unit = self.UnitClass("teststring")
        unit.target = '''quoted 'string' and "string"'''

        assert (
            str(unit)
            == """<string name="teststring">quoted 'string' and "string"</string>\n"""
        )


class TestMOKOResourceFile(test_monolingual.TestMonolingualStore):
    StoreClass = aresource.MOKOResourceFile

    def test_plural(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plural name="vms_num_visitors">
        <item quantity="one">%d visitor</item>
        <item quantity="other">%d visitors</item>
    </plural>
</resources>
"""
        store = self.StoreClass()
        store.parse(content.encode())
        assert store.units[0].target == multistring(["%d visitor", "%d visitors"])
        store = self.StoreClass()
        store.targetlanguage = "zh-rHK"
        store.parse(content.encode())
        store.units[0].target = "%d 訪客"
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plural name="vms_num_visitors">
        <item quantity="other">%d 訪客</item>
    </plural>
</resources>
"""
        )


class TestCMPResourceUnit(TestMOKOResourceUnit):
    UnitClass = aresource.CMPResourceUnit


class TestCMPResourceFile(test_monolingual.TestMonolingualStore):
    StoreClass = aresource.CMPResourceFile

    def test_quote_escape(self) -> None:
        store = self.StoreClass()

        # Test parsing
        store.parse(
            r"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="cloud_changelogScreen_title">Bienvenue dans la version préliminaire de l\'application Efento Cloud</string>
</resources>
""".encode()
        )
        assert len(store.units) == 1
        assert (
            store.units[0].target
            == r"Bienvenue dans la version préliminaire de l\'application Efento Cloud"
        )

        # Test updating
        store.units[
            0
        ].target = "Bienvenue dans la version préliminaire de l'application"
        assert (
            bytes(store).decode()
            == r"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="cloud_changelogScreen_title">Bienvenue dans la version préliminaire de l'application</string>
</resources>
"""
        )
        store.units[
            0
        ].target = r"Bienvenue dans la version préliminaire de l\'application"
        assert (
            bytes(store).decode()
            == r"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="cloud_changelogScreen_title">Bienvenue dans la version préliminaire de l\\'application</string>
</resources>
"""
        )



from io import BytesIO

from translate.convert import moz2po, prop2po
from translate.storage import properties

from . import test_convert


class TestMoz2PO:
    @staticmethod
    def moz2po(propsource, proptemplate=None):
        """Helper that converts Mozilla .properties source to po source without requiring files."""
        inputfile = BytesIO(propsource.encode())
        inputprop = properties.propfile(inputfile, personality="mozilla")
        convertor = prop2po.prop2po(personality="mozilla")
        if proptemplate:
            templatefile = BytesIO(proptemplate.encode())
            templateprop = properties.propfile(templatefile, personality="mozilla")
            outputpo = convertor.mergestore(templateprop, inputprop)
        else:
            outputpo = convertor.convertstore(inputprop)
        return outputpo

    def test_duplicate_locations(self) -> None:
        """
        Test that duplicate locations in Mozilla properties files don't cause
        AttributeError. This tests the fix for the Lithuanian Firefox recovery issue.
        """
        # Template file with duplicate keys (like in Lithuanian Firefox files)
        templateprop = """css.property.one=First value
css.property.one=Duplicate value
css.property.two=Second value
"""
        # Translation file with duplicate keys
        translatedprop = """css.property.one=Primera valor
css.property.one=Valor duplicado
css.property.two=Segunda valor
"""
        # This should not raise AttributeError: 'NoneType' object has no attribute 'getid'
        pofile = self.moz2po(translatedprop, templateprop)

        # Verify we got results (not testing exact structure, just that it didn't crash)
        assert len(pofile.units) > 1
        # At least one unit should have the location 'css.property.one'
        locations = []
        for unit in pofile.units:
            if not unit.isheader():
                locations.extend(unit.getlocations())
        assert "css.property.one" in locations
        assert "css.property.two" in locations


class TestMoz2POCommand(test_convert.TestConvertCommand, TestMoz2PO):
    """Tests running actual moz2po commands on files."""

    convertmodule = moz2po
    defaultoptions = {"progress": "none"}

    expected_options = [
        "--duplicates=DUPLICATESTYLE",
        "-P, --pot",
        "-t TEMPLATE, --template=TEMPLATE",
    ]



from io import BytesIO

from translate.misc.multistring import multistring
from translate.storage import php

from . import test_monolingual


def test_php_escaping_single_quote() -> None:
    """
    Test the helper escaping functions for 'single quotes'.

    The tests are built mostly from examples from the PHP
    `string type definition <http://www.php.net/manual/en/language.types.string.php#language.types.string.syntax.single>`_.
    """
    # Decoding - PHP -> Python
    assert (
        php.phpdecode(r"\'") == r"'"
    )  # To specify a literal single quote, escape it with a backslash (\).
    assert php.phpdecode(r'"') == r'"'
    assert (
        php.phpdecode(r"\\'") == r"\'"
    )  # To specify a literal backslash before a single quote, or at the end of the string, double it (\\)
    assert (
        php.phpdecode(r"\x") == r"\x"
    )  # Note that attempting to escape any other character will print the backslash too.
    assert php.phpdecode(r"\t") == r"\t"
    assert php.phpdecode(r"\n") == r"\n"
    assert php.phpdecode(r"this is a simple string") == r"this is a simple string"
    assert (
        php.phpdecode(
            """You can also have embedded newlines in
strings this way as it is
okay to do"""
        )
        == """You can also have embedded newlines in
strings this way as it is
okay to do"""
    )
    assert (
        php.phpdecode(r"This will not expand: \n a newline")
        == r"This will not expand: \n a newline"
    )
    assert (
        php.phpdecode(r'Arnold once said: "I\'ll be back"')
        == r'''Arnold once said: "I'll be back"'''
    )
    assert php.phpdecode(r"You deleted C:\\*.*?") == r"You deleted C:\*.*?"
    assert php.phpdecode(r"You deleted C:\*.*?") == r"You deleted C:\*.*?"
    assert (
        php.phpdecode(r"\117\143\164\141\154") == r"\117\143\164\141\154"
    )  # We don't handle Octal like " does
    assert php.phpdecode(r"\x48\x65\x78") == r"\x48\x65\x78"  # Don't handle Hex either
    # Should implement for false interpretation of double quoted data.
    # Encoding - Python -> PHP
    assert (
        php.phpencode(r"'") == r"\'"
    )  # To specify a literal single quote, escape it with a backslash (\).
    assert (
        php.phpencode(r"\'") == r"\\'"
    )  # To specify a literal backslash before a single quote, or at the end of the string, double it (\\)
    assert php.phpencode(r'"') == r'"'
    assert (
        php.phpencode(r"\x") == r"\x"
    )  # Note that attempting to escape any other character will print the backslash too.
    assert php.phpencode(r"\t") == r"\t"
    assert php.phpencode(r"\n") == r"\n"
    assert (
        php.phpencode(
            r"""String with
newline"""
        )
        == r"""String with
newline"""
    )
    assert (
        php.phpencode(r"This will not expand: \n a newline")
        == r"This will not expand: \n a newline"
    )
    assert (
        php.phpencode(r'''Arnold once said: "I'll be back"''')
        == r'''Arnold once said: "I\'ll be back"'''
    )
    assert php.phpencode(r"You deleted C:\*.*?") == r"You deleted C:\*.*?"


def test_php_escaping_double_quote() -> None:
    """Test the helper escaping functions for 'double quotes'."""
    # Decoding - PHP -> Python
    assert php.phpdecode("'", quotechar='"') == "'"  # we do nothing with single quotes
    assert (
        php.phpdecode(r"\n", quotechar='"') == "\n"
    )  # See table of escaped characters
    assert (
        php.phpdecode(r"\r", quotechar='"') == "\r"
    )  # See table of escaped characters
    assert (
        php.phpdecode(r"\t", quotechar='"') == "\t"
    )  # See table of escaped characters
    assert (
        php.phpdecode(r"\v", quotechar='"') == "\v"
    )  # See table of escaped characters
    assert (
        php.phpdecode(r"\f", quotechar='"') == "\f"
    )  # See table of escaped characters
    assert (
        php.phpdecode(r"\\", quotechar='"') == "\\"
    )  # See table of escaped characters
    # assert php.phpdecode(r"\$", quotechar='"') == "$"      # See table of escaped characters - this may cause confusion with actual variables in roundtripping
    assert (
        php.phpdecode(r"\$", quotechar='"') == "\\$"
    )  # Just to check that we don't unescape this
    assert php.phpdecode(r"\"", quotechar='"') == '"'  # See table of escaped characters
    assert (
        php.phpdecode(r"\117\143\164\141\154", quotechar='"') == "Octal"
    )  # Octal: \[0-7]{1,3}
    assert (
        php.phpdecode(r"\x48\x65\x78", quotechar='"') == "Hex"
    )  # Hex: \x[0-9A-Fa-f]{1,2}
    assert php.phpdecode(r"\117\\c\164\141\154", quotechar='"') == r"O\ctal"  # Mixed
    # Decoding - special examples
    assert (
        php.phpdecode(r"Don't escape me here\'s", quotechar='"')
        == r"Don't escape me here\'s"
    )  # See bug #589
    assert (
        php.phpdecode("Line1\nLine2") == "Line1\nLine2"
    )  # Preserve newlines in multiline messages
    assert php.phpdecode("Line1\r\nLine2") == "Line1\r\nLine2"  # DOS PHP files
    # Encoding - Python -> PHP
    assert php.phpencode("'", quotechar='"') == "'"
    assert (
        php.phpencode("\n", quotechar='"') == "\n"
    )  # See table of escaped characters - we leave newlines unescaped so that we can try best to preserve pretty printing. See bug 588
    assert (
        php.phpencode("\r", quotechar='"') == r"\r"
    )  # See table of escaped characters
    assert (
        php.phpencode("\t", quotechar='"') == r"\t"
    )  # See table of escaped characters
    assert (
        php.phpencode("\v", quotechar='"') == r"\v"
    )  # See table of escaped characters
    assert (
        php.phpencode("\f", quotechar='"') == r"\f"
    )  # See table of escaped characters
    assert (
        php.phpencode(r"\\", quotechar='"') == r"\\"
    )  # See table of escaped characters
    # assert php.phpencode("\$", quotechar='"') == "$"      # See table of escaped characters - this may cause confusion with actual variables in roundtripping
    assert (
        php.phpencode(r"\$", quotechar='"') == r"\$"
    )  # Just to check that we don't unescape this
    assert php.phpencode('"', quotechar='"') == r"\""
    assert (
        php.phpencode(r"Don't escape me here\'s", quotechar='"')
        == r"Don't escape me here\'s"
    )  # See bug #589


class TestPhpUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = php.phpunit

    def test_difficult_escapes(self) -> None:
        pass


class TestPhpFile(test_monolingual.TestMonolingualStore):
    StoreClass = php.phpfile

    @staticmethod
    def phpparse(phpsource):
        """Helper that parses php source without requiring files."""
        dummyfile = BytesIO(phpsource.encode())
        return php.phpfile(dummyfile)

    def phpregen(self, phpsource):
        """Helper that converts php source to phpfile object and back."""
        return bytes(self.phpparse(phpsource)).decode("utf-8")

    def test_simpledefinition(self) -> None:
        """Checks that a simple php definition is parsed correctly."""
        phpsource = """$lang['mediaselect'] = 'Bestand selective';"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang['mediaselect']"
        assert phpunit.source == "Bestand selective"

    def test_simpledefinition_source(self) -> None:
        """Checks that a simple php definition can be regenerated as source."""
        phpsource = """<?php
$lang['mediaselect'] = 'Bestand selective';"""
        phpregen = self.phpregen(phpsource)
        assert f"{phpsource}\n" == phpregen

    def test_spaces_in_name(self) -> None:
        """Check that spaces in the array name doesn't throw us off."""
        phpsource = """$lang[ 'mediaselect' ] = 'Bestand selective';"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang[ 'mediaselect' ]"
        assert phpunit.source == "Bestand selective"

    def test_comment_definition(self) -> None:
        """Check that comments are fully preserved."""
        phpsource = """/*
 * Comment line 1
 * Comment line 2
 */
$foo = "bar";
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "$foo"
        assert phpunit.source == "bar"
        assert phpunit._comments == [
            """/*
 * Comment line 1
 * Comment line 2
 */"""
        ]

    def test_comment_blocks(self) -> None:
        """Check that we don't process name value pairs in comment blocks."""
        phpsource = """/*
 * $lang[0] = "Blah";
 * $lang[1] = "Bluh";
 */
$lang[2] = "Yeah";
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang[2]"
        assert phpunit.source == "Yeah"

    def test_comment_output(self) -> None:
        """Check that linebreaks and spacing is preserved when comments are output."""
        phpsource = """/*
 * Comment line 1
 * Comment line 2
 */
$foo = 'bar';
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert str(phpunit) == phpsource

    def test_comment_add(self) -> None:
        """Check that comments are actually added."""
        phpsource = """/* NOTE 1 */
$foo = 'bar';
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "$foo"
        assert phpunit.source == "bar"
        assert phpunit._comments == ["/* NOTE 1 */"]
        # Replace existing notes
        phpunit.addnote("/* NOTE 2 */", None, "replace")
        # Set unknown note
        phpunit.addnote("/* NOTE 2 */", "unknown")
        assert phpunit._comments == ["/* NOTE 2 */"]
        assert str(phpunit) == phpsource.replace("NOTE 1", "NOTE 2")

    def test_multiline(self) -> None:
        """Check that we preserve newlines in a multiline message."""
        phpsource = """$lang['multiline'] = "Line1%sLine2";"""
        # Try DOS and Unix and make sure the output has the same
        for lineending in ("\n", "\r\n"):
            phpfile = self.phpparse(phpsource % lineending)
            assert len(phpfile.units) == 1
            phpunit = phpfile.units[0]
            assert phpunit.name == "$lang['multiline']"
            assert phpunit.source == f"Line1{lineending}Line2"

    def test_parsing_arrays(self) -> None:
        """Parse the array syntax."""
        phpsource = """$lang = %s(
         'item1' => 'value1',
         'item2' => 'value2',
      );"""
        for arrayfn in ["array", "Array", "ARRAY"]:
            phpfile = self.phpparse(phpsource % arrayfn)
            assert len(phpfile.units) == 2
            phpunit = phpfile.units[0]
            assert phpunit.name == "$lang->'item1'"
            assert phpunit.source == "value1"

    def test_parsing_array_no_array_syntax(self) -> None:
        """Parse the array syntax."""
        phpsource = """global $_LANGPDF;
        $_LANGPDF = array();
        $_LANGPDF['PDF065ab3a28ca4f16f55f103adc7d0226f'] = 'Delivery';
        """
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "$_LANGPDF['PDF065ab3a28ca4f16f55f103adc7d0226f']"
        assert phpunit.source == "Delivery"

    def test_parsing_arrays_keys_with_spaces(self) -> None:
        """Ensure that our identifiers can have spaces. Bug #1683."""
        phpsource = """$lang = array(
         'item 1' => 'value1',
         'item 2' => 'value2',
      );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang->'item 1'"
        assert phpunit.source == "value1"

    def test_parsing_arrays_keys_with_quotes(self) -> None:
        """Ensure that our identifiers get escaped. Bug #5348."""
        phpsource = r"""<?php
return [
    'item\'s 1' => 'value1',
    'item "2"' => 'value2',
];
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == "return[]->'item's 1'"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "return[]->'item \"2\"'"
        assert phpunit.source == "value2"

        assert bytes(phpfile).decode() == phpsource

    def test_parsing_arrays_keys_with_number_as_value(self) -> None:
        """Ensure that our identifiers can have numbers as value. Bug #5355."""
        phpsource = r"""<?php
return [
    1 => 'value1',
    2 => 'value2',
    '3' => 'value3',
    '4\'' => 'value4',
    '5"' => 'value5',
    '6\"' => 'value6',
];
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 6
        phpunit = phpfile.units[0]
        assert phpunit.name == "return[]->1"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "return[]->2"
        assert phpunit.source == "value2"
        phpunit = phpfile.units[2]
        assert phpunit.name == "return[]->'3'"
        assert phpunit.source == "value3"
        phpunit = phpfile.units[3]
        assert phpunit.name == "return[]->'4''"
        assert phpunit.source == "value4"
        phpunit = phpfile.units[4]
        assert phpunit.name == "return[]->'5\"'"
        assert phpunit.source == "value5"
        phpunit = phpfile.units[5]
        assert phpunit.name == "return[]->'6\\\"'"
        assert phpunit.source == "value6"

        assert bytes(phpfile).decode() == phpsource

    def test_parsing_arrays_non_textual(self) -> None:
        """Don't break on non-textual data. Bug #1684."""
        phpsource = """$lang = array(
         'item 1' => 'value1',
         'item 2' => false,
         'item 3' => 'value3',
      );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang->'item 3'"
        assert phpunit.source == "value3"

    def test_parsing_simple_define(self) -> None:
        """Parse simple define syntax."""
        phpsource = """define("_FINISH", "Rematar");
define('_POSTEDON', 'Enviado o');"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define("_FINISH"'
        assert phpunit.source == "Rematar"
        phpunit = phpfile.units[1]
        assert phpunit.name == "define('_POSTEDON'"
        assert phpunit.source == "Enviado o"

    def test_parsing_simple_define_with_spaces_before_key(self) -> None:
        """Parse simple define syntax with spaces before key."""
        phpsource = """define( "_FINISH", "Rematar");
define( '_CM_POSTED', 'Enviado');"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define( "_FINISH"'
        assert phpunit.source == "Rematar"
        phpunit = phpfile.units[1]
        assert phpunit.name == "define( '_CM_POSTED'"
        assert phpunit.source == "Enviado"

    def test_parsing_define_spaces_after_equal_delimiter(self) -> None:
        """Parse define syntax with spaces after the equal delimiter."""
        phpsource = """define("_RELOAD",       "Recargar");
define('_CM_POSTED',    'Enviado');"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define("_RELOAD"'
        assert phpunit.source == "Recargar"
        phpunit = phpfile.units[1]
        assert phpunit.name == "define('_CM_POSTED'"
        assert phpunit.source == "Enviado"

    def test_parsing_define_spaces_after_equal_delimiter_and_before_key(self) -> None:
        """
        Parse define syntax with spaces after the equal delimiter as well
        before the key.
        """
        phpsource = """define( "_FINISH",       "Rematar");
define(  '_UPGRADE_CHARSET',    'Upgrade charset');"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define( "_FINISH"'
        assert phpunit.source == "Rematar"
        phpunit = phpfile.units[1]
        assert phpunit.name == "define(  '_UPGRADE_CHARSET'"
        assert phpunit.source == "Upgrade charset"

    def test_parsing_define_no_spaces_after_equal_delimiter(self) -> None:
        """Parse define syntax without spaces after the equal delimiter."""
        phpsource = """define("_POSTEDON","Enviado o");
define('_UPGRADE_CHARSET','Upgrade charset');"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define("_POSTEDON"'
        assert phpunit.source == "Enviado o"
        phpunit = phpfile.units[1]
        assert phpunit.name == "define('_UPGRADE_CHARSET'"
        assert phpunit.source == "Upgrade charset"

    def test_parsing_define_no_spaces_after_equaldel_but_before_key(self) -> None:
        """
        Parse define syntax without spaces after the equal delimiter but
        with spaces before the key.
        """
        phpsource = """define( "_FINISH","Rematar");
define( '_CM_POSTED','Enviado');"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define( "_FINISH"'
        assert phpunit.source == "Rematar"
        phpunit = phpfile.units[1]
        assert phpunit.name == "define( '_CM_POSTED'"
        assert phpunit.source == "Enviado"

    def test_parsing_define_entries_with_quotes(self) -> None:
        """Parse define syntax for entries with quotes."""
        phpsource = r"""define('_SETTINGS_COOKIEPREFIX', 'Prefixo da "cookie"');
define('_YOUR_USERNAME', 'O seu nome de usuario: "cookie"');
define("_REGISTER", "Register <a href=\"register.php\">here</a>");"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 3
        phpunit = phpfile.units[0]
        assert phpunit.name == "define('_SETTINGS_COOKIEPREFIX'"
        assert phpunit.source == 'Prefixo da "cookie"'
        phpunit = phpfile.units[1]
        assert phpunit.name == "define('_YOUR_USERNAME'"
        assert phpunit.source == 'O seu nome de usuario: "cookie"'
        phpunit = phpfile.units[2]
        assert phpunit.name == 'define("_REGISTER"'
        assert phpunit.source == 'Register <a href="register.php">here</a>'

    def test_parsing_define_comments_at_entry_line_end(self) -> None:
        """Parse define syntax with comments at the end of the entry line."""
        phpsource = """define("_POSTEDON", "Enviado o");// Keep this short
define('_CM_POSTED', 'Enviado'); // Posted date"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define("_POSTEDON"'
        assert phpunit.source == "Enviado o"
        assert phpunit._comments == ["// Keep this short"]
        phpunit = phpfile.units[1]
        assert phpunit.name == "define('_CM_POSTED'"
        assert phpunit.source == "Enviado"
        assert phpunit._comments == ["// Posted date"]

    def test_parsing_define_double_slash_comments_before_entries(self) -> None:
        """Parse define syntax with double slash comments before the entries."""
        phpsource = """// Keep this short
define("_FINISH", "Rematar");

// This means it was published
// It appears besides posts
define('_CM_POSTED', 'Enviado');"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define("_FINISH"'
        assert phpunit.source == "Rematar"
        assert phpunit._comments == ["// Keep this short"]
        phpunit = phpfile.units[1]
        assert phpunit.name == "define('_CM_POSTED'"
        assert phpunit.source == "Enviado"
        assert phpunit._comments == [
            "// This means it was published",
            "// It appears besides posts",
        ]

    def test_parsing_define_spaces_before_end_delimiter(self) -> None:
        """Parse define syntax with spaces before the end delimiter."""
        phpsource = """define("_POSTEDON", "Enviado o");
define("_FINISH", "Rematar"     );
define("_RELOAD", "Recargar");"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 3
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define("_POSTEDON"'
        assert phpunit.source == "Enviado o"
        phpunit = phpfile.units[1]
        assert phpunit.name == 'define("_FINISH"'
        assert phpunit.source == "Rematar"
        phpunit = phpfile.units[2]
        assert phpunit.name == 'define("_RELOAD"'
        assert phpunit.source == "Recargar"

    def test_parsing_simpledefinition_spaces_before_end_delimiter(self) -> None:
        """Parse simple definition syntax with spaces before the end delimiter."""
        phpsource = """$month_jan = 'Jan';
$month_feb = 'Feb'  ;
$month_mar = 'Mar';"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 3
        phpunit = phpfile.units[0]
        assert phpunit.name == "$month_jan"
        assert phpunit.source == "Jan"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$month_feb"
        assert phpunit.source == "Feb"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$month_mar"
        assert phpunit.source == "Mar"

    def test_parsing_arrays_no_trailing_comma(self) -> None:
        """
        Parse the array syntax where we don't have a trailing comma.

        Bug #1685
        """
        phpsource = """$lang = array(
         'item1' => 'value1',
         'item2' => 'value2'
      );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang->'item1'"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang->'item2'"
        assert phpunit.source == "value2"

    def test_parsing_arrays_space_before_comma(self) -> None:
        """
        Parse the array syntax with spaces before the comma.

        Bug #1898
        """
        phpsource = """$lang = array(
         'item1' => 'value1',
         'item2' => 'value2' ,
        );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang->'item1'"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang->'item2'"
        assert phpunit.source == "value2"

    def test_parsing_arrays_with_space_before_array_declaration(self) -> None:
        """
        Parse the array syntax with spaces before the array declaration.

        Bug #2646
        """
        phpsource = """$lang = array   (
         'item1' => 'value1',
         'item2' => 'value2',
        );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang->'item1'"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang->'item2'"
        assert phpunit.source == "value2"

    def test_parsing_arrays_declared_in_a_single_line(self) -> None:
        """
        Parse an array declared in a single line.

        Bug #3629
        """
        phpsource = """$lang = array( 'item1' => 'value1', 'item2' => 'value2', 'item3' => 'value3' );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 3
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang->'item1'"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang->'item2'"
        assert phpunit.source == "value2"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$lang->'item3'"
        assert phpunit.source == "value3"

    def test_parsing_arrays_with_no_keys(self) -> None:
        """Parse an array with no keys."""
        phpsource = """<?php
$days = array(
    'Sunday',
    'Monday',
    'Tuesday',
    'Wednesday',
    'Thursday',
    'Friday',
    "Saturday",
);
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 7
        phpunit = phpfile.units[0]
        assert phpunit.name == "$days->[]"
        assert phpunit.source == "Sunday"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$days->[]"
        assert phpunit.source == "Monday"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$days->[]"
        assert phpunit.source == "Tuesday"
        phpunit = phpfile.units[3]
        assert phpunit.name == "$days->[]"
        assert phpunit.source == "Wednesday"
        phpunit = phpfile.units[4]
        assert phpunit.name == "$days->[]"
        assert phpunit.source == "Thursday"
        phpunit = phpfile.units[5]
        assert phpunit.name == "$days->[]"
        assert phpunit.source == "Friday"
        phpunit = phpfile.units[6]
        assert phpunit.name == "$days->[]"
        assert phpunit.source == "Saturday"
        assert bytes(phpfile).decode() == phpsource

    def test_parsing_arrays_with_no_keys_assigned_to_array(self) -> None:
        """Parse an array with no keys assigned to another array."""
        phpsource = """<?php
$messages['days_short'] = array(
    'Sun',
    'Mon',
    'Tue',
    'Wed',
    'Thu',
    "Fri",
    'Sat',
);
"""
        phpfile = self.phpparse(phpsource)
        assert bytes(phpfile).decode() == phpsource
        assert len(phpfile.units) == 7
        phpunit = phpfile.units[0]
        assert phpunit.name == "$messages['days_short']->[]"
        assert phpunit.source == "Sun"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$messages['days_short']->[]"
        assert phpunit.source == "Mon"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$messages['days_short']->[]"
        assert phpunit.source == "Tue"
        phpunit = phpfile.units[3]
        assert phpunit.name == "$messages['days_short']->[]"
        assert phpunit.source == "Wed"
        phpunit = phpfile.units[4]
        assert phpunit.name == "$messages['days_short']->[]"
        assert phpunit.source == "Thu"
        phpunit = phpfile.units[5]
        assert phpunit.name == "$messages['days_short']->[]"
        assert phpunit.source == "Fri"
        phpunit = phpfile.units[6]
        assert phpunit.name == "$messages['days_short']->[]"
        assert phpunit.source == "Sat"

    def test_parsing_nested_arrays_with_no_keys(self) -> None:
        """Parse a nested array with no keys."""
        phpsource = """$lang = array(array("key" => "value"));"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang->[]->'key'"
        assert phpunit.source == "value"

    def test_assignment_in_line_where_multiline_comment_ends(self) -> None:
        """Check parsing assignment in same line a multiline comment ends."""
        phpsource = """<?php
/*
   Multi-line
   comment
*/ $messages['help'] = 'Help';
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "$messages['help']"
        assert phpunit.source == "Help"

    def test_parsing_arrays_using_short_array_syntax(self) -> None:
        """Parse short array syntax.  Bug #3626."""
        phpsource = """<?php
$lang = [
    'item1' => 'value1',
    'item2' => 'value2',
];
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang[]->'item1'"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang[]->'item2'"
        assert phpunit.source == "value2"
        assert bytes(phpfile).decode() == phpsource

    def test_parsing_nested_arrays(self) -> None:
        """Parse the nested array syntax. Bug #2240."""
        phpsource = """$app_list_strings = array(
            'Mailbox' => 'Mailbox',
            'moduleList' => array(
                'Home' => 'Home',
                'Contacts' => 'Contacts',
                'Accounts' => 'Accounts',
            ),
            'FAQ' => 'FAQ',
        );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 5
        phpunit = phpfile.units[0]
        assert phpunit.name == "$app_list_strings->'Mailbox'"
        assert phpunit.source == "Mailbox"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$app_list_strings->'moduleList'->'Home'"
        assert phpunit.source == "Home"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$app_list_strings->'moduleList'->'Contacts'"
        assert phpunit.source == "Contacts"
        phpunit = phpfile.units[3]
        assert phpunit.name == "$app_list_strings->'moduleList'->'Accounts'"
        assert phpunit.source == "Accounts"
        phpunit = phpfile.units[4]
        assert phpunit.name == "$app_list_strings->'FAQ'"
        assert phpunit.source == "FAQ"

    def test_parsing_nested_arrays_with_space_before_array_declaration(self) -> None:
        """Parse the nested array syntax with whitespace before the array declaration."""
        phpsource = """$app_list_strings = array  (
            'Mailbox' => 'Mailbox',
            'moduleList' => array  (
                'Home' => 'Home',
                'Contacts' => 'Contacts',
                'Accounts' => 'Accounts',
            ),
            'tools' => array  (
                'Pen' => 'Pen',
                'Brush' => 'Brush',
                'Pencil' => 'Pencil',
            ),
            'FAQ' => 'FAQ',
        );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 8
        phpunit = phpfile.units[0]
        assert phpunit.name == "$app_list_strings->'Mailbox'"
        assert phpunit.source == "Mailbox"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$app_list_strings->'moduleList'->'Home'"
        assert phpunit.source == "Home"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$app_list_strings->'moduleList'->'Contacts'"
        assert phpunit.source == "Contacts"
        phpunit = phpfile.units[3]
        assert phpunit.name == "$app_list_strings->'moduleList'->'Accounts'"
        assert phpunit.source == "Accounts"
        phpunit = phpfile.units[4]
        assert phpunit.name == "$app_list_strings->'tools'->'Pen'"
        assert phpunit.source == "Pen"
        phpunit = phpfile.units[5]
        assert phpunit.name == "$app_list_strings->'tools'->'Brush'"
        assert phpunit.source == "Brush"
        phpunit = phpfile.units[6]
        assert phpunit.name == "$app_list_strings->'tools'->'Pencil'"
        assert phpunit.source == "Pencil"
        phpunit = phpfile.units[7]
        assert phpunit.name == "$app_list_strings->'FAQ'"
        assert phpunit.source == "FAQ"

    def test_parsing_unnamed_nested_arrays(self) -> None:
        """Parse the unnamed nested array."""
        phpsource = """return array  (
            'name1' => 'target1',
            'list1' => array(
                'l1' => 'target_l1_1',
                'l2' => 'target_l1_2',
                'l3' => 'target_l1_3',
            ),
            'list2' => array(
                'l1' => 'target_l2_1',
                'l2' => 'target_l2_2',
                'l3' => 'target_l2_3',
            ),
            'name2' => 'target2',
        );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 8
        phpunit = phpfile.units[0]
        assert phpunit.name == "return->'name1'"
        assert phpunit.source == "target1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "return->'list1'->'l1'"
        assert phpunit.source == "target_l1_1"
        phpunit = phpfile.units[2]
        assert phpunit.name == "return->'list1'->'l2'"
        assert phpunit.source == "target_l1_2"
        phpunit = phpfile.units[3]
        assert phpunit.name == "return->'list1'->'l3'"
        assert phpunit.source == "target_l1_3"
        phpunit = phpfile.units[4]
        assert phpunit.name == "return->'list2'->'l1'"
        assert phpunit.source == "target_l2_1"

    def test_parsing_nested_arrays_with_array_declaration_in_next_line(self) -> None:
        """
        Parse the nested array syntax with array declaration in the next line.

        Bug #2647
        """
        phpsource = """$lang = array(
            'item1' => 'value1',
            'newsletter_frequency_dom' =>
                array(
                    'Weekly' => 'Weekly',
                ),
            'item2' => 'value2',
        );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 3
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang->'item1'"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang->'newsletter_frequency_dom'->'Weekly'"
        assert phpunit.source == "Weekly"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$lang->'item2'"
        assert phpunit.source == "value2"

    def test_parsing_array_with_newline_after_delimiter(self) -> None:
        """
        Parse array with newline between key and value.

        Bug #3628.
        """
        phpsource = """$lang = array(
            'item1' => 'value1',
            'frequency' =>
                'Weekly',
            'item2' => 'value2',
        );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 3
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang->'item1'"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang->'frequency'"
        assert phpunit.source == "Weekly"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$lang->'item2'"
        assert phpunit.source == "value2"

    def test_parsing_nested_arrays_with_blank_entries(self) -> None:
        """Parse the nested array syntax with blank entries. Bug #2648."""
        phpsource = """<?php
$lang = array(
    'item1' => 'value1',
    'newsletter_frequency_dom' => array(
        '' => '',
        'Weekly' => 'Weekly',
    ),
    'item2' => 'value2',
);
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 4
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang->'item1'"
        assert phpunit.source == "value1"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang->'newsletter_frequency_dom'->''"
        assert phpunit.source == ""
        phpunit = phpfile.units[2]
        assert phpunit.name == "$lang->'newsletter_frequency_dom'->'Weekly'"
        assert phpunit.source == "Weekly"
        phpunit = phpfile.units[3]
        assert phpunit.name == "$lang->'item2'"
        assert phpunit.source == "value2"
        assert bytes(phpfile).decode() == phpsource

    def test_slashstar_in_string(self) -> None:
        """Ignore the /* comment delimiter when it is part of a string.  Bug #3627."""
        phpsource = """$definition['key'] = 'Value /* value continued';
      $lang = array(
         'somekey' => 'Some value',
         'otherkey' => 'Other value /* continued',
         'thirdkey' => 'Third value',
      );"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 4
        phpunit = phpfile.units[0]
        assert phpunit.name == "$definition['key']"
        assert phpunit.source == "Value /* value continued"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang->'somekey'"
        assert phpunit.source == "Some value"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$lang->'otherkey'"
        assert phpunit.source == "Other value /* continued"
        phpunit = phpfile.units[3]
        assert phpunit.name == "$lang->'thirdkey'"
        assert phpunit.source == "Third value"

    def test_parsing_simple_heredoc_syntax(self) -> None:
        """Parse the heredoc syntax. Bug #2611."""
        phpsource = """$month_jan = 'Jan';
$lang_register_approve_email = <<<EOT
A new user with the username "{USER_NAME}" has registered in your gallery.

In order to activate the account, you need to click on the link below.

<a href="{ACT_LINK}">{ACT_LINK}</a>
EOT;

$foobar = <<<FOOBAR
Simple example
FOOBAR;

$month_mar = 'Mar';
        """
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 4
        phpunit = phpfile.units[0]
        assert phpunit.name == "$month_jan"
        assert phpunit.source == "Jan"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang_register_approve_email"
        assert (
            phpunit.source
            == 'A new user with the username "{USER_NAME}" has registered in your gallery.\n\nIn order to activate the account, you need to click on the link below.\n\n<a href="{ACT_LINK}">{ACT_LINK}</a>'
        )
        phpunit = phpfile.units[2]
        assert phpunit.name == "$foobar"
        assert phpunit.source == "Simple example"
        phpunit = phpfile.units[3]
        assert phpunit.name == "$month_mar"
        assert phpunit.source == "Mar"

    def test_simpledefinition_after_define(self) -> None:
        """Check that a simple definition after define is parsed correctly."""
        phpsource = """define("_FINISH", "Rematar");
$lang['mediaselect'] = 'Bestand selective';"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == 'define("_FINISH"'
        assert phpunit.source == "Rematar"
        phpunit = phpfile.units[1]
        assert phpunit.name == "$lang['mediaselect']"
        assert phpunit.source == "Bestand selective"

    def test_quotes(self) -> None:
        phpsource = """<?php
$txt[\'DISPLAYEDINFOS\'] = "<a href=\\"__PARAM2__\\">Modificar...</a>";
"""
        phpfile = self.phpparse(phpsource)
        assert bytes(phpfile).decode() == phpsource

    def test_concatenation(self) -> None:
        """Check that concatenating strings and variables is parsed correctly."""
        phpsource = """
$lang['mediaselect'] = "Really \\\"something\\\"" . $variable . "\\\"something else\\\"";
$messages['welcome'] = 'Welcome ' . $name . '!';
$messages['greeting'] = 'Hi ' . $name;
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 3
        phpunit = phpfile.units[0]
        assert phpunit.name == "$lang['mediaselect']"
        assert phpunit.source == 'Really "something"$variable"something else"'
        phpunit = phpfile.units[1]
        assert phpunit.name == "$messages['welcome']"
        assert phpunit.source == "Welcome $name!"
        phpunit = phpfile.units[2]
        assert phpunit.name == "$messages['greeting']"
        assert phpunit.source == "Hi $name"

    def test_serialize(self) -> None:
        phpsource = """<?php
# Comment 1
define("_FINISH", 'Rematar');
// Comment 2
define("_START", "Zacni");
/* Comment 3 */
$lang['mediaselect'] = 'Bestand selective';
$texts = array(
    /*
     * Multiline comment 4
     */
    'other' => 'Andere',
    'nested' => array(
        'and' => 'Und',
        'second' => 'Zweite',
    ),
);
"""
        phpfile = self.phpparse(phpsource)
        assert bytes(phpfile).decode() == phpsource

    def test_space_before_comma(self) -> None:
        """
        Check that spacing before comma or semicolon doesn't break parser.

        See #1898
        """
        phpsource = """
        $english = array(
            'item:site'  =>  "Sítios",
            'login'  =>  "Entrar" ,
            'loginok'  =>  "Entrou com sucesso." ,
        );
        $month_jan = 'Jan';
        $month_feb = 'Feb'  ;
        $month_mar = 'Mar';
        define("_SEARCH","Search");
        define("_LOGIN","Login"  );
        define("_POLLS","Polls");
        """
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 9

    def test_equals_in_id(self) -> None:
        """
        Check that equals in id doesn't break parser.

        See #1929
        """
        phpsource = """
        $strings['key = value'] = 'Message';
        """
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        assert phpfile.units[0].source == "Message"
        assert phpfile.units[0].name == "$strings['key = value']"

    def test_comma_in_string(self) -> None:
        """
        Check that comma in string doesn't break parser.

        See #3608
        """
        phpsource = """
        $t = array(
            'key' => ' text (comment **),**',
        );
        """
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        assert phpfile.units[0].source == " text (comment **),**"
        assert phpfile.units[0].name == "$t->'key'"

    def test_nowdoc(self) -> None:
        """Check parsing nowdoc strings."""
        phpsource = """$str = <<<'EOD'
Example of string
spanning multiple lines
using nowdoc syntax.
EOD;
        """
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        assert (
            phpfile.units[0].source
            == """Example of string
spanning multiple lines
using nowdoc syntax."""
        )
        assert phpfile.units[0].name == "$str"

    def test_plain_concatenation(self) -> None:
        """Check parsing concatenated strings."""
        phpsource = """$str = 'Concatenated' . ' ' . 'string';
        $arr['x'] = 'Concatenated' . ' ' . 'string';
        """
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        assert phpfile.units[0].source == "Concatenated string"
        assert phpfile.units[0].name == "$str"
        assert phpfile.units[1].source == "Concatenated string"
        assert phpfile.units[1].name == "$arr['x']"

    def test_array_keys(self) -> None:
        """Check parsing different array keys."""
        phpsource = """
        $arr = [
            '1234' => 'First',
            1234 => 'Second',
            '12' . '45' => 'Third',
        ];
        """
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 3
        assert phpfile.units[0].source == "First"
        assert phpfile.units[0].name == "$arr[]->'1234'"
        assert phpfile.units[1].source == "Second"
        assert phpfile.units[1].name == "$arr[]->1234"
        assert phpfile.units[2].source == "Third"
        assert phpfile.units[2].name == "$arr[]->'1245'"

    def test_double_var(self) -> None:
        """Checks that a double $ is handled correctly."""
        phpsource = """$$lang['mediaselect'] = 'Bestand selective';"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "$$lang['mediaselect']"
        assert phpunit.source == "Bestand selective"

    def test_return_array(self) -> None:
        phpsource = """<?php
return array(
    'peach' => 'pesca',
);
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "return->'peach'"
        assert phpunit.source == "pesca"
        assert bytes(phpfile).decode() == phpsource

    def test_return_array_with_spaces(self) -> None:
        """Test that return array with spaces before 'array' is handled correctly."""
        phpsource = """<?php
return  array(
    'peach' => 'pesca',
);
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        # The unit name should be "return->'peach'", NOT "return  array->'peach'"
        assert phpunit.name == "return->'peach'"
        assert phpunit.source == "pesca"
        # The output should normalize spacing
        assert bytes(phpfile) == b"<?php\nreturn array(\n    'peach' => 'pesca',\n);\n"

    def test_return_array_with_comments(self) -> None:
        """Test that return array with comments preserves comments correctly."""
        phpsource = """<?php

return array(

    /*
    |--------------------------------------------------------------------------
    | String1
    |--------------------------------------------------------------------------
    |
    */

  'string1'    => 'This is string1',
);
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "return->'string1'"
        assert phpunit.source == "This is string1"
        # Check that comments are preserved
        assert "String1" in phpunit.getnotes("developer")

    def test_return_array_short(self) -> None:
        phpsource = """<?php
return [
    'peach' => 'pesca',
];
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.name == "return[]->'peach'"
        assert phpunit.source == "pesca"
        assert bytes(phpfile).decode() == phpsource
        phpunit.source = "ryba"
        assert bytes(phpfile).decode() != phpsource
        phpunit.source = "pesca"
        assert bytes(phpfile).decode() == phpsource

    def test_return_array_short_quotes(self) -> None:
        phpsource = r"""<?php
return [
    'peach' => "foo \"pesca\"",
];
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 1
        phpunit = phpfile.units[0]
        assert phpunit.escape_type == '"'
        assert phpunit.name == "return[]->'peach'"
        assert phpunit.source == 'foo "pesca"'
        assert bytes(phpfile).decode() == phpsource
        phpunit.source = "ryba"
        assert bytes(phpfile) != phpsource
        phpunit.source = 'foo "pesca"'
        assert bytes(phpfile).decode() == phpsource

    def test_addunit(self) -> None:
        expected = """<?php
$key = 'first';
$sec = 'second';
"""
        store = self.StoreClass()

        unit = self.StoreClass.UnitClass("first")
        unit.setid("key")
        store.addunit(unit)
        unit = self.StoreClass.UnitClass("second")
        unit.setid("$sec")
        store.addunit(unit)
        assert bytes(store).decode() == expected


class TestLaravelPhpUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = php.LaravelPHPUnit


class TestLaravelPhpFile(test_monolingual.TestMonolingualStore):
    StoreClass = php.LaravelPHPFile

    def phpparse(self, phpsource):
        """Helper that parses php source without requiring files."""
        dummyfile = BytesIO(phpsource.encode())
        return self.StoreClass(dummyfile)

    def test_plurals(self) -> None:
        phpsource = r"""<?php
return [
    'welcome' => 'Welcome to our application',
    'apples' => 'There is one apple|There are many apples',
];
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        phpunit = phpfile.units[0]
        assert phpunit.name == "return[]->'welcome'"
        assert phpunit.source == "Welcome to our application"
        phpunit = phpfile.units[1]
        assert phpunit.name == "return[]->'apples'"
        assert phpunit.source == multistring(
            ["There is one apple", "There are many apples"]
        )
        assert bytes(phpfile).decode() == phpsource
        phpunit.source = multistring(["There is an apple", "There are many apples"])
        assert bytes(phpfile).decode() == phpsource.replace("one apple", "an apple")

    def test_key_stripping(self) -> None:
        """Test that Laravel PHP files strip the return prefix from keys."""
        phpsource = r"""<?php
return [
    'welcome' => 'Welcome to our application',
    'apples' => 'There is one apple|There are many apples',
];
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        # Check that getid() returns keys without the return[]-> prefix
        assert phpfile.units[0].getid() == "welcome"
        assert phpfile.units[1].getid() == "apples"
        # Check that getlocations() also returns stripped keys
        assert phpfile.units[0].getlocations() == ["welcome"]
        assert phpfile.units[1].getlocations() == ["apples"]

    def test_key_stripping_array_syntax(self) -> None:
        """Test key stripping works with array() syntax."""
        phpsource = r"""<?php
return array(
    'welcome' => 'Welcome',
    'goodbye' => 'Goodbye',
);
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        assert phpfile.units[0].getid() == "welcome"
        assert phpfile.units[1].getid() == "goodbye"

    def test_key_stripping_numeric_keys(self) -> None:
        """Test key stripping works with numeric keys."""
        phpsource = r"""<?php
return [
    1 => 'One',
    2 => 'Two',
];
"""
        phpfile = self.phpparse(phpsource)
        assert len(phpfile.units) == 2
        assert phpfile.units[0].getid() == "1"
        assert phpfile.units[1].getid() == "2"

    def test_roundtrip_short_array(self) -> None:
        """Test round trip serialization with short array syntax."""
        phpsource = r"""<?php
return [
    'welcome' => 'Welcome to our application',
    'apples' => 'There is one apple|There are many apples',
];
"""
        phpfile = self.phpparse(phpsource)
        assert bytes(phpfile).decode() == phpsource

    def test_roundtrip_array_syntax(self) -> None:
        """Test round trip serialization with array() syntax."""
        phpsource = r"""<?php
return array(
    'welcome' => 'Welcome',
    'goodbye' => 'Goodbye',
);
"""
        phpfile = self.phpparse(phpsource)
        assert bytes(phpfile).decode() == phpsource

    def test_setid_preserves_structure(self) -> None:
        """Test that setid() preserves the return array structure."""
        phpsource = r"""<?php
return [
    'welcome' => 'Welcome',
];
"""
        phpfile = self.phpparse(phpsource)
        unit = phpfile.units[0]

        # Change the key
        unit.setid("greeting")

        # Check that getid() returns the clean key
        assert unit.getid() == "greeting"

        # Check that serialization produces valid PHP with return structure
        output = bytes(phpfile).decode()
        assert "return [" in output
        assert "'greeting' => 'Welcome'" in output
        assert "return[]->" not in output  # Should not leak internal structure

    def test_setid_with_array_syntax(self) -> None:
        """Test that setid() preserves array() syntax."""
        phpsource = r"""<?php
return array(
    'welcome' => 'Welcome',
);
"""
        phpfile = self.phpparse(phpsource)
        unit = phpfile.units[0]

        # Change the key
        unit.setid("greeting")

        # Check that getid() returns the clean key
        assert unit.getid() == "greeting"

        # Check that serialization produces valid PHP with return array() structure
        output = bytes(phpfile).decode()
        assert "return array(" in output
        assert "'greeting' => 'Welcome'" in output

    def test_addunit_with_setid(self) -> None:
        """Test creating new units programmatically with setid()."""
        phpfile = self.StoreClass()

        unit1 = phpfile.addsourceunit("Welcome")
        unit1.setid("welcome")

        unit2 = phpfile.addsourceunit("Goodbye")
        unit2.setid("goodbye")

        # Check IDs
        assert unit1.getid() == "welcome"
        assert unit2.getid() == "goodbye"

        # Check serialization produces valid Laravel PHP structure
        output = bytes(phpfile).decode()
        expected = """<?php
return array(
    'welcome' => 'Welcome',
    'goodbye' => 'Goodbye',
);
"""
        assert output == expected

    def test_add_unit_to_short_array_file(self) -> None:
        """Test adding a new unit to a file with short array syntax []."""
        phpsource = r"""<?php
return [
    'existing' => 'Existing message',
];
"""
        phpfile = self.phpparse(phpsource)

        # Add a new unit
        new_unit = phpfile.addsourceunit("New message")
        new_unit.setid("new_key")

        # Check the new unit has correct ID
        assert new_unit.getid() == "new_key"

        # Check serialization produces valid PHP with consistent [] syntax
        output = bytes(phpfile).decode()
        expected = """<?php
return [
    'existing' => 'Existing message',
    'new_key' => 'New message',
];
"""
        assert output == expected

    def test_add_unit_to_array_function_file(self) -> None:
        """Test adding a new unit to a file with array() syntax."""
        phpsource = r"""<?php
return array(
    'existing' => 'Existing message',
);
"""
        phpfile = self.phpparse(phpsource)

        # Add a new unit
        new_unit = phpfile.addsourceunit("New message")
        new_unit.setid("new_key")

        # Check the new unit has correct ID
        assert new_unit.getid() == "new_key"

        # Check serialization produces valid PHP with consistent array() syntax
        output = bytes(phpfile).decode()
        expected = """<?php
return array(
    'existing' => 'Existing message',
    'new_key' => 'New message',
);
"""
        assert output == expected

    def test_add_multiple_units_to_short_array(self) -> None:
        """Test adding multiple units to a file with short array syntax."""
        phpsource = r"""<?php
return [
    'first' => 'First',
];
"""
        phpfile = self.phpparse(phpsource)

        # Add multiple new units
        unit2 = phpfile.addsourceunit("Second")
        unit2.setid("second")

        unit3 = phpfile.addsourceunit("Third")
        unit3.setid("third")

        # Check serialization produces valid PHP
        output = bytes(phpfile).decode()
        expected = """<?php
return [
    'first' => 'First',
    'second' => 'Second',
    'third' => 'Third',
];
"""
        assert output == expected

    def test_add_multiple_units_to_array_function(self) -> None:
        """Test adding multiple units to a file with array() syntax."""
        phpsource = r"""<?php
return array(
    'first' => 'First',
);
"""
        phpfile = self.phpparse(phpsource)

        # Add multiple new units
        unit2 = phpfile.addsourceunit("Second")
        unit2.setid("second")

        unit3 = phpfile.addsourceunit("Third")
        unit3.setid("third")

        # Check serialization produces valid PHP
        output = bytes(phpfile).decode()
        expected = """<?php
return array(
    'first' => 'First',
    'second' => 'Second',
    'third' => 'Third',
);
"""
        assert output == expected

    def test_numeric_keys_no_quotes(self) -> None:
        """Test that numeric keys (including negative) are not quoted."""
        phpfile = self.StoreClass()

        # Test positive integer
        unit1 = phpfile.addsourceunit("Message 1")
        unit1.setid("1")

        # Test negative integer
        unit2 = phpfile.addsourceunit("Message 2")
        unit2.setid("-1")

        # Test zero
        unit3 = phpfile.addsourceunit("Message 3")
        unit3.setid("0")

        output = bytes(phpfile).decode()
        expected = """<?php
return array(
    1 => 'Message 1',
    -1 => 'Message 2',
    0 => 'Message 3',
);
"""
        assert output == expected

    def test_empty_string_key(self) -> None:
        """Test that empty string keys are properly quoted."""
        phpfile = self.StoreClass()

        unit = phpfile.addsourceunit("Empty key message")
        unit.setid("")

        output = bytes(phpfile).decode()
        expected = """<?php
return array(
    '' => 'Empty key message',
);
"""
        assert output == expected

    def test_setid_with_non_string_value(self) -> None:
        """Test that setid handles non-string values gracefully."""
        phpfile = self.StoreClass()

        # Test with integer
        unit1 = phpfile.addsourceunit("Message")
        unit1.setid(123)

        output = bytes(phpfile).decode()
        expected = """<?php
return array(
    123 => 'Message',
);
"""
        assert output == expected



from translate.lang import factory


def test_punctranslate() -> None:
    """Tests that we can translate punctuation."""
    language = factory.getlanguage("nqo")
    assert language.punctranslate("") == ""
    assert language.punctranslate("abc efg") == "abc efg"
    assert language.punctranslate("abc efg.") == "abc efg."
    assert language.punctranslate("abc efg!") == "abc efg߹"
    assert language.punctranslate("abc, efg; d?") == "abc߸ efg؛ d؟"
    # See https://github.com/translate/translate/issues/1819
    assert language.punctranslate("It is called “abc”") == "It is called ”abc“"


def test_sentences() -> None:
    """Tests basic functionality of sentence segmentation."""
    language = factory.getlanguage("nqo")
    sentences = language.sentences("")
    assert sentences == []

    # this text probably does not make sense, I just copied it from Firefox
    # translation and added some punctuation marks
    sentences = language.sentences(
        "ߡߍ߲ ߠߎ߬ ߦߋ߫ ߓߊ߯ߙߊ߫ ߟߊ߫ ߢߐ߲߮ ߝߍ߬ ߞߊ߬ ߓߟߐߟߐ ߟߊߞߊ߬ߣߍ߲ ߕߏ߫. ߖߊ߬ߡߊ ߣߌ߫ ߓߍ߯ ߛߊ߬ߥߏ ߘߐ߫."
    )
    print(sentences)
    assert sentences == [
        "ߡߍ߲ ߠߎ߬ ߦߋ߫ ߓߊ߯ߙߊ߫ ߟߊ߫ ߢߐ߲߮ ߝߍ߬ ߞߊ߬ ߓߟߐߟߐ ߟߊߞߊ߬ߣߍ߲ ߕߏ߫.",
        "ߖߊ߬ߡߊ ߣߌ߫ ߓߍ߯ ߛߊ߬ߥߏ ߘߐ߫.",
    ]
    sentences = language.sentences(
        "ߡߍ߲ ߠߎ߬ ߦߋ߫ ߓߊ߯ߙߊ߫ ߟߊ߫ ߢߐ߲߮ ߝߍ߬ ߞߊ߬ ߓߟߐߟߐ ߟߊߞߊ߬ߣߍ߲ ߕߏ߫? ߖߊ߬ߡߊ ߣߌ߫ ߓߍ߯ ߛߊ߬ߥߏ ߘߐ߫."
    )
    print(sentences)
    assert sentences == [
        "ߡߍ߲ ߠߎ߬ ߦߋ߫ ߓߊ߯ߙߊ߫ ߟߊ߫ ߢߐ߲߮ ߝߍ߬ ߞߊ߬ ߓߟߐߟߐ ߟߊߞߊ߬ߣߍ߲ ߕߏ߫?",
        "ߖߊ߬ߡߊ ߣߌ߫ ߓߍ߯ ߛߊ߬ߥߏ ߘߐ߫.",
    ]



from lxml import etree

from translate.misc.multistring import multistring
from translate.storage import poxliff

from . import test_xliff


class TestPOXLIFFUnit(test_xliff.TestXLIFFUnit):
    UnitClass = poxliff.PoXliffUnit

    def test_plurals(self) -> None:
        """Tests that plurals are handled correctly."""
        unit = self.UnitClass(multistring(["Cow", "Cows"]))
        print(type(unit.source))
        print(repr(unit.source))
        assert isinstance(unit.source, multistring)
        assert unit.source.strings == ["Cow", "Cows"]
        assert unit.source == "Cow"

        unit.target = ["Koei", "Koeie"]
        assert isinstance(unit.target, multistring)
        assert unit.target.strings == ["Koei", "Koeie"]
        assert unit.target == "Koei"

        unit.target = ["Sk\u00ear", "Sk\u00eare"]
        assert isinstance(unit.target, multistring)
        assert unit.target.strings == ["Sk\u00ear", "Sk\u00eare"]
        assert unit.target.strings == ["Sk\u00ear", "Sk\u00eare"]
        assert unit.target == "Sk\u00ear"

    def test_ids(self) -> None:
        """Tests that ids are assigned correctly, especially for plurals."""
        unit = self.UnitClass("gras")
        assert not unit.getid()
        unit.setid("4")
        assert unit.getid() == "4"

        unit = self.UnitClass(multistring(["shoe", "shoes"]))
        assert not unit.getid()
        unit.setid("20")
        assert unit.getid() == "20"
        assert unit.units[1].getid() == "20[1]"

        unit.target = ["utshani", "uutshani", "uuutshani"]
        assert unit.getid() == "20"
        assert unit.units[1].getid() == "20[1]"

    def test_setsource_preserves_child_ids(self) -> None:
        """Tests that setsource propagates group ID to new child trans-units."""
        unit = self.UnitClass(multistring(["cow", "cows"]))
        unit.setid("L_PLU.TEST")
        assert unit.units[0].xmlelement.get("id") == "L_PLU.TEST[0]"
        assert unit.units[1].xmlelement.get("id") == "L_PLU.TEST[1]"

        # Resetting source should propagate IDs to new children
        unit.source = multistring(["cow", "cows", "cows"])
        assert unit.xmlelement.get("id") == "L_PLU.TEST"
        assert unit.units[0].xmlelement.get("id") == "L_PLU.TEST[0]"
        assert unit.units[1].xmlelement.get("id") == "L_PLU.TEST[1]"
        assert unit.units[2].xmlelement.get("id") == "L_PLU.TEST[2]"

    def test_setsource_singular_to_plural_rebuilds_group(self) -> None:
        """Tests that switching from singular to plural rebuilds the XML node."""
        unit = self.UnitClass("cow")
        unit.setid("L_PLU.TEST")
        unit.target = "vache"
        unit.addnote("translator note")

        unit.source = multistring(["cow", "cows"])

        assert unit.hasplural()
        assert unit.xmlelement.tag == unit.namespaced("group")
        assert unit.xmlelement.get("restype") == "x-gettext-plurals"
        assert len(unit.units) == 2
        assert [child.tag for child in unit.xmlelement] == [
            unit.namespaced("note"),
            unit.namespaced("trans-unit"),
            unit.namespaced("trans-unit"),
        ]
        assert unit.units[0].xmlelement.get("id") == "L_PLU.TEST[0]"
        assert unit.units[1].xmlelement.get("id") == "L_PLU.TEST[1]"
        assert unit.target.strings == ["vache", ""]

    def test_setsource_nonplural_on_plural_group(self) -> None:
        """
        Tests that setting a non-plural source on a plural group doesn't
        add <source> to the group element.
        """
        unit = self.UnitClass(multistring(["cow", "cows"]))
        unit.setid("L_PLU.TEST")

        # Setting a single string source should update child units, not group
        unit.source = "updated"
        assert unit.hasplural()
        # Should NOT have <source> directly on the group
        ns = unit.namespaced("source")
        group_sources = [c for c in unit.xmlelement if c.tag == ns]
        assert len(group_sources) == 0
        # Child units should have the updated source
        assert unit.units[0].source == "updated"
        assert unit.units[1].source == "updated"

    def test_setsource_nonplural_removes_stale_group_source(self) -> None:
        """
        Tests that setting a non-plural source on a plural group also
        removes any stale <source> element on the group.
        """
        unit = self.UnitClass(multistring(["cow", "cows"]))
        unit.setid("L_PLU.TEST")
        # Manually inject a stale <source> on the group (simulating old buggy data)
        stale_source = etree.SubElement(unit.xmlelement, unit.namespaced("source"))
        stale_source.text = "stale"
        ns = unit.namespaced("source")
        assert len([c for c in unit.xmlelement if c.tag == ns]) == 1

        # Setting non-plural source should clean up the stale <source>
        unit.source = "updated"
        group_sources = [c for c in unit.xmlelement if c.tag == ns]
        assert len(group_sources) == 0
        assert unit.units[0].source == "updated"
        assert unit.units[1].source == "updated"

    def test_settarget_none_on_plural(self) -> None:
        """
        Tests that setting target to None on a plural unit doesn't create
        empty target elements.
        """
        unit = self.UnitClass(multistring(["cow", "cows"]))
        unit.setid("L_PLU.TEST")
        unit.target = multistring(["vache", "vaches"])
        assert unit.target.strings == ["vache", "vaches"]

        # Setting target to None should clear targets
        unit.target = None
        for child_unit in unit.units:
            assert child_unit.target is None

    def test_construction_no_empty_targets(self) -> None:
        """
        Tests that constructing a plural unit doesn't add empty <target>
        elements to child trans-units.
        """
        unit = self.UnitClass(multistring(["cow", "cows"]))
        ns = unit.namespaced("target")
        for child_unit in unit.units:
            targets = list(child_unit.xmlelement.iterchildren(ns))
            assert len(targets) == 0

    def test_setsource_plural_without_targets_keeps_target_nodes_absent(self) -> None:
        """Tests that resetting plural source doesn't create empty targets."""
        unit = self.UnitClass(multistring(["cow", "cows"]))
        unit.setid("L_PLU.TEST")

        unit.source = multistring(["cow", "cows", "calves"])

        ns = unit.namespaced("target")
        for child_unit in unit.units:
            targets = list(child_unit.xmlelement.iterchildren(ns))
            assert len(targets) == 0


class TestPOXLIFFfile(test_xliff.TestXLIFFfile):
    StoreClass = poxliff.PoXliffFile
    xliffskeleton = """<?xml version="1.0" ?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
  <file original="filename.po" source-language="en-US" datatype="po">
    <body>
        %s
    </body>
  </file>
</xliff>"""

    def test_parse(self) -> None:
        minixlf = (
            self.xliffskeleton
            % """<group restype="x-gettext-plurals">
        <trans-unit id="1[0]" xml:space="preserve">
            <source>cow</source>
            <target>inkomo</target>
        </trans-unit>
        <trans-unit id="1[1]" xml:space="preserve">
            <source>cows</source>
            <target>iinkomo</target>
        </trans-unit>
</group>"""
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert len(xlifffile.units) == 1
        assert xlifffile.units[0].hasplural()
        assert xlifffile.translate("cow") == "inkomo"
        assert xlifffile.units[0].source == "cow"
        assert xlifffile.units[0].source.strings == ["cow", "cows"]
        assert xlifffile.units[0].target.strings == ["inkomo", "iinkomo"]

    def test_parse_plural_alpha_id(self) -> None:
        minixlf = (
            self.xliffskeleton
            % """<group restype="x-gettext-plurals">
        <trans-unit id="test[0]" xml:space="preserve">
            <source>cow</source>
            <target>inkomo</target>
        </trans-unit>
        <trans-unit id="test[1]" xml:space="preserve">
            <source>cows</source>
            <target>iinkomo</target>
        </trans-unit>
</group>"""
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert len(xlifffile.units) == 1
        assert xlifffile.units[0].hasplural()
        assert xlifffile.translate("cow") == "inkomo"
        assert xlifffile.units[0].source == "cow"
        assert xlifffile.units[0].source.strings == ["cow", "cows"]
        assert xlifffile.units[0].target.strings == ["inkomo", "iinkomo"]

    def test_notes(self) -> None:
        minixlf = (
            self.xliffskeleton
            % """<group restype="x-gettext-plurals">
        <trans-unit id="1[0]" xml:space="preserve">
            <source>cow</source>
            <target>inkomo</target>
<note from="po-translator">Zulu translation of program ABC</note>
<note from="developer">azoozoo come back!</note>
        </trans-unit>
        <trans-unit id="1[1]" xml:space="preserve">
            <source>cows</source>
            <target>iinkomo</target>
<note from="po-translator">Zulu translation of program ABC</note>
<note from="developer">azoozoo come back!</note>
        </trans-unit>
</group>"""
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert (
            xlifffile.units[0].getnotes()
            == "Zulu translation of program ABC\nazoozoo come back!"
        )
        assert xlifffile.units[0].getnotes("developer") == "azoozoo come back!"
        assert (
            xlifffile.units[0].getnotes("po-translator")
            == "Zulu translation of program ABC"
        )

    def test_plural(self) -> None:
        minixlf = (
            self.xliffskeleton
            % """
        <group id="1238108068" restype="x-gettext-plurals">
                <trans-unit id="1238108068[0]" xml:space="preserve">
                        <source>This field must contain at least {0,number} character</source>
                </trans-unit>
                <trans-unit id="1238108068[1]" xml:space="preserve">
                        <source>This field must contain at least {0,number} characters</source>
                </trans-unit>
        </group>
            """
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]
        assert unit.source.strings == [
            "This field must contain at least {0,number} character",
            "This field must contain at least {0,number} characters",
        ]
        assert unit.target == ["", ""]

    def test_parse_plural_many_forms(self) -> None:
        """Tests parsing plural groups with more than 6 forms (e.g. Arabic)."""
        forms = [
            f'<trans-unit id="ar[{i}]" xml:space="preserve">'
            f"<source>form{i}</source>"
            f"<target>target{i}</target>"
            f"</trans-unit>"
            for i in range(7)
        ]
        minixlf = self.xliffskeleton % (
            '<group id="ar" restype="x-gettext-plurals">'
            + "\n".join(forms)
            + "</group>"
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]
        assert unit.hasplural()
        assert len(unit.units) == 7
        assert unit.source.strings == [f"form{i}" for i in range(7)]

    def test_parse_nonplural_bracketed_numeric_id(self) -> None:
        """Tests that standalone units with bracketed numeric IDs are kept."""
        minixlf = (
            self.xliffskeleton
            % """<trans-unit id="menu[12]" xml:space="preserve">
        <source>Menu item</source>
        <target>Polozka menu</target>
      </trans-unit>"""
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]
        assert not unit.hasplural()
        assert unit.xmlelement.get("id") == "menu[12]"
        assert unit.source == "Menu item"
        assert unit.target == "Polozka menu"



from zipfile import ZIP_DEFLATED, ZipFile

import pytest

from translate.misc.zipfile_helpers import (
    MAX_ARCHIVE_MEMBER_SIZE,
    MAX_ARCHIVE_TOTAL_SIZE,
)
from translate.storage import idml


def test_open_idml_rejects_large_members(tmp_path) -> None:
    archive = tmp_path / "large.idml"
    with ZipFile(archive, "w", compression=ZIP_DEFLATED) as zip_file:
        zip_file.writestr(
            "Stories/Story_u1.xml",
            b"A" * (MAX_ARCHIVE_MEMBER_SIZE + 1),
        )

    with pytest.raises(ValueError, match="too large after decompression"):
        idml.open_idml(archive)


def test_copy_idml_rejects_large_members(tmp_path) -> None:
    archive = tmp_path / "large-template.idml"
    output = tmp_path / "out.idml"
    with ZipFile(archive, "w", compression=ZIP_DEFLATED) as zip_file:
        zip_file.writestr("Stories/Story_u1.xml", b"A")
        zip_file.writestr("Resources/Graphic.ai", b"B" * (MAX_ARCHIVE_MEMBER_SIZE + 1))

    with (
        ZipFile(archive, "r") as input_zip,
        ZipFile(output, "w") as output_zip,
        pytest.raises(ValueError, match="too large after decompression"),
    ):
        idml.copy_idml(input_zip, output_zip, {"Stories/Story_u1.xml"})


def test_copy_idml_allows_large_total_size_when_members_are_bounded(tmp_path) -> None:
    archive = tmp_path / "large-total-template.idml"
    output = tmp_path / "out-large-total.idml"
    member_size = MAX_ARCHIVE_TOTAL_SIZE // 3
    with ZipFile(archive, "w", compression=ZIP_DEFLATED) as zip_file:
        zip_file.writestr("Stories/Story_u1.xml", b"A")
        zip_file.writestr("Resources/Graphic-1.ai", b"B" * member_size)
        zip_file.writestr("Resources/Graphic-2.ai", b"C" * member_size)
        zip_file.writestr("Resources/Graphic-3.ai", b"D" * member_size)
        zip_file.writestr("Resources/Graphic-4.ai", b"E" * member_size)

    with ZipFile(archive, "r") as input_zip, ZipFile(output, "w") as output_zip:
        idml.copy_idml(input_zip, output_zip, {"Stories/Story_u1.xml"})

    with ZipFile(output, "r") as output_zip:
        assert output_zip.read("Resources/Graphic-4.ai") == b"E" * member_size



import os

from translate.convert import xliff2odf

from . import test_convert


class TestXLIFF2ODFommand(test_convert.TestConvertCommand):
    """Tests running actual xliff2odf commands on files."""

    convertmodule = xliff2odf
    expected_options = [
        "-t TEMPLATE, --template=TEMPLATE",
    ]

    def test_convert(self) -> None:
        xliffsource = """<?xml version="1.0" ?>
    <xliff version="1.1">
      <file original="filename.po" source-language="en-US" datatype="po">
        <body>
         <trans-unit xml:space="preserve" id="1" approved="yes">
           <source>Hello, world!</source>
           <target state="translated">Ahoj světe!</target>
         </trans-unit>
        </body>
      </file>
    </xliff>"""
        self.create_testfile("simple.xlf", xliffsource)
        self.run_command(
            i="simple.xlf",
            o="simple.odt",
            template=os.path.join(os.path.dirname(__file__), "test.odt"),
        )
